"""
Constants extracted from:ETSI TS 103 300-3 V2.2.1 (2023-02)
Table 17: Parameters for VAM generation triggering (clause 6.4)

Parametrs: minimumSafeLateralDistance and minimumSafeLongitudinalDistance are
    not defined here, as they are not static(they depend on the VRU speed)
"""
T_GENVAMMIN = 100  # ms
T_GENVAMMAX = 5000  # ms
T_CHECKVAMGEN = T_GENVAMMIN  # ms Shall be equal to or less than T_GenvamMin
T_GENVAM_DCC = T_GENVAMMIN  # ms T_GenvamMin ≤ T_Genvam_DCC ≤ T_GenvamMax
MINREFERENCEPOINTPOSITIONCHANGETHRESHOLD = 4
MINGROUNDSPEEDCHANGETHRESHOLD = 0.5
MINGROUNDVELOCITYORIENTATIONCHANGETHRESHOLD = 4
MINTRAJECTORYINTERCEPTIONPROBCHANGETHRESHOLD = 10
NUMSKIPVAMSFORREDUNDANCYMITIGATION = 2  # Value can range from 2-10
MINCLUSTERDISTANCECHANGETHRESHOLD = 2
MINIMUMSAFEVERTICALDISTANCE = 5

"""
From ETSI TS 103 300-2 V2.1.1 (2020-05), page 47;
The VRU Basic Service shall interact with the VRU profile management entity in
    the management layer to learn
whether the ITS-S has the VRU role activated.

TODO: Create VRU Profile Mangement

Since it's not yet created the values will be declared here
"""

VRU_PROFILE = {
    "Type": "Cyclist",
    "Speed": 20,
    "TransmissionRange": 70,
    "Environment": "Urban",
    "WeightClass": "High",
    "TrajectoryAmbiguity": "Medium",
    "ClusterSize": 1,
}

"""
ETSI TS 103 300-3 V2.2.1 (2023-02) - 5.4.1 VRU clustering functional overview
States: The support of the clustering function is optional in the VBS for all
VRU profiles.
however the same document in section C.2.3 Protocol data, states;
The VRU Basic Service (VBS) stores at least the following information for the
VAM originating ITS-S operation:
VAM generation time;
• ITS-S position as included in VAM;
• ITS-S speed as included in VAM;
• ITS-S heading as included in VAM;
• VRU role;
• VRU profile;
• VBS cluster state.
VRU role, VRU profile nad VBS cluster state will be hardcoded (VRU_Profile already is).
"""

VRU_ROLE = "VRU_ROLE_ON"
VRU_CLUSTER_STATE = "VRU-ACTIVE-STANDALONE"



from enum import Enum

from .exceptions import DecodeError


class M(Enum):
    """
    M field of GNAddress as described in ETSI EN 302 636-4-1 V1.4.1 (2020-01)

    ...

    Attributes
    ----------
    GN_UNICAST :
        GN Unicast Address
    GN_MULTICAST :
        GN Multicast Address
    """
    GN_UNICAST = 0
    GN_MULTICAST = 1

    def encode_to_address(self) -> int:
        """
        Encodes M to int for GN address

        Returns
        -------
        int
            Encoded M
        """
        return (self.value << 7) << 8*7


class ST(Enum):
    """
    ST field of GNAddress as described in ETSI EN 302 636-4-1 V1.4.1 (2020-01)

    ...

    Attributes
    ----------
    UNKNOWN :
        Unknown
    PEDESTRIAN :
        Pedestrianremaining
    CYCLIST :
        Cyclist
    MOPED :
        Moped
    MOTORCYCLE :
        Motorcycle
    PASSENGER_CAR :
        Passenger Car
    BUS :
        Bus
    LIGHT_TRUCK :
        Light Truck
    HEAVY_TRUCK :
        Heavy Truck
    TRAILER :
        Trailer
    SPECIAL_VEHICLE :
        Special Vehicle
    TRAM :
        Tram
    ROAD_SIDE_UNIT :
        Road Side Unit
    """
    UNKNOWN = 0
    PEDESTRIAN = 1
    CYCLIST = 2
    MOPED = 3
    MOTORCYCLE = 4
    PASSENGER_CAR = 5
    BUS = 6
    LIGHT_TRUCK = 7
    HEAVY_TRUCK = 8
    TRAILER = 9
    SPECIAL_VEHICLE = 10
    TRAM = 11
    ROAD_SIDE_UNIT = 12

    def encode_to_address(self) -> int:
        """
        Encodes ST to int for GN address

        Returns
        int
        bytes
            Encoded ST
        """
        return (self.value << 3) << 8*7


class InvalidMIDLength(Exception):
    """
    Exception for invalid MID length
    """


class MID:
    """
    MID field of GNAddress as described in ETSI EN 302 636-4-1 V1.4.1 (2020-01)

    ...

    Attributes
    ----------
    MID : bytes
        MID field of GNAddress: LL_ADDR
    """

    def __init__(self, mid: bytes):
        if len(mid) != 6:
            raise InvalidMIDLength("MID must be 6 bytes long")
        self.mid: bytes = mid

    def encode_to_address(self) -> int:
        """
        Encodes MID to int for GN address

        Returns
        -------
        int
            Encoded MID for GN address
        """
        return int.from_bytes(b'\x00\x00'+self.mid, byteorder='big')


class GNAddress:
    """
    GeoNetworking Address as described in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 6.3

    ...

    Attributes
    ----------
    M : M
        M field of GNAddress
    ST : ST
        ST field of GNAddress
    MID : bytes
        MID field of GNAddress: LL_ADDR
    """

    def __init__(self):
        self.m: M = M.GN_UNICAST
        self.st: ST = ST.UNKNOWN
        self.mid: MID = MID(b'\x00\x00\x00\x00\x00\x00')

    def set_m(self, m: M):
        """
        Sets M field of GNAddress

        Parameters
        ----------
        m : M
            M field of GNAddress
        """
        self.m = m

    def set_st(self, st: ST):
        """
        Sets ST field of GNAddress

        Parameters
        ----------
        st : ST
            ST field of GNAddress
        """
        self.st = st

    def set_mid(self, mid: MID):
        """
        Sets MID field of GNAddress

        Parameters
        ----------
        mid : MID
            MID field of GNAddress
        """
        self.mid = mid

    def encode(self) -> bytes:
        """
        Encodes GNAddress to bytes

        Returns
        -------
        bytes
            Encoded GNAddress
        """
        return (self.m.encode_to_address() | self.st.encode_to_address() | self.mid.encode_to_address()).to_bytes(
            8, byteorder='big')

    def encode_to_int(self) -> int:
        """
        Encodes GNAddress to int

        Returns
        -------
        int
            Encoded GNAddress
        """
        return self.m.encode_to_address() | self.st.encode_to_address() | self.mid.encode_to_address()

    def decode(self, data: bytes):
        """
        Decodes GNAddress from bytes
        """
        if len(data) < 8:
            raise DecodeError("GNAddress must be 8 bytes long")
        self.m = M((data[0] & 0x80) >> 7)
        self.st = ST((data[0] & 0x78) >> 3)
        self.mid = MID(data[2:8])

    def __eq__(self, __o: object) -> bool:
        """
        Checks if two GNAddress are equal.

        Just compares the MAC address inside the MID field

        Parameters
        ----------
        __o : object
            Object to compare
        """
        return self.mid.mid == __o.mid.mid

    def __str__(self) -> str:
        return f"GNAddress(M={self.m}, ST={self.st}, MID={self.mid})"



import asn1tools
from .security_asn1 import SECURITY_ASN1_DESCRIPTIONS


class SecurityCoder:
    """
    Class for encoding/decoding Security Headers, Certificates and other Security related formats
    Follows the ETSI TS 103 097 V2.1.1 (2021-10) standard.

    Attributes
    ----------
    asn_coder : asn1tools.Coder
        ASN.1 coder.
    """

    def __init__(self) -> None:
        """
        Initialize the Security Coder.
        """
        self.asn_coder = asn1tools.compile_string(
            SECURITY_ASN1_DESCRIPTIONS, codec="oer"
        )

    def encode_etsi_ts_103097_certificate(self, certificate: dict) -> bytes:
        """
        Encode a EtsiTs103097Certificate.

        Parameters
        ----------
        certificate : dict
            The certificate to encode.

        Returns
        -------
        bytes
            The encoded certificate.
        """
        return self.asn_coder.encode("EtsiTs103097Certificate", certificate)

    def decode_etsi_ts_103097_certificate(self, certificate: bytes) -> dict:
        """
        Decode a EtsiTs103097Certificate.

        Parameters
        ----------
        certificate : bytes
            The certificate to decode.

        Returns
        -------
        dict
            The decoded certificate.
        """
        return self.asn_coder.decode("EtsiTs103097Certificate", certificate)

    def encode_etsi_ts_103097_data_signed(self, data: dict) -> bytes:
        """
        Encode a EtsiTs103097Data-Signed.

        Parameters
        ----------
        data : dict
            The data to encode.

        Returns
        -------
        bytes
            The encoded data.
        """
        return self.asn_coder.encode("EtsiTs103097Data", data)

    def decode_etsi_ts_103097_data_signed(self, data: bytes) -> dict:
        """
        Decode a EtsiTs103097Data-Signed.

        Parameters
        ----------
        data : bytes
            The data to decode.

        Returns
        -------
        dict
            The decoded data.
        """
        return self.asn_coder.decode("EtsiTs103097Data", data)

    def encode_to_be_signed_data(self, data: dict) -> bytes:
        """
        Encode a ToBeSignedData.

        Parameters
        ----------
        data : dict
            The data to encode.

        Returns
        -------
        bytes
            The encoded data.
        """
        return self.asn_coder.encode("ToBeSignedData", data)

    def encode_ToBeSignedCertificate(self, data: dict) -> bytes:
        """
        Encode a ToBeSignedCertificate.

        Parameters
        ----------
        data : dict
            The data to encode.

        Returns
        -------
        bytes
            The encoded data.
        """
        return self.asn_coder.encode("ToBeSignedCertificate", data)



from __future__ import annotations
import threading
import time

from .ldm_classes import Location, Filter
from .database import DataBase
from .ldm_maintenance import LDMMaintenance


class LDMMaintenanceThread(LDMMaintenance):
    """
    Class that inherits from LDMMaintenance (class specified in the ETSI ETSI EN 302 895 V1.1.1 (2014-09).

    This class is used to run the maintenance of the LDM in a seperate thread. This is done to ensure that the LDM
    is always up to date and that the data is always valid. This is done by overriding the add_provider_data method.
    """
    def __init__(
        self,
        area_of_maintenance: Location,
        data_base: DataBase,
        stop_event: threading.Event,
    ) -> None:
        super().__init__(area_of_maintenance, data_base)
        self.data_containers_lock = threading.Lock()
        self.stop_event = stop_event
        self.ldm_maintenance_thread = threading.Thread(target=self.run, daemon=True)
        self.ldm_maintenance_thread.start()

    def run(self) -> None:
        while not self.stop_event.is_set():
            time.sleep(1)
            # Continually monitor the data containers and delete expired data
            self.collect_trash()

    def add_provider_data(self, data) -> int:
        with self.data_containers_lock:
            doc_id = super().add_provider_data(data)

        return doc_id

    def get_provider_data(self, data_object_id) -> dict:
        with self.data_containers_lock:
            provider_data = super().get_provider_data(data_object_id)

        return provider_data

    def update_provider_data(self, data_object_id, data_object) -> None:
        with self.data_containers_lock:
            super().update_provider_data(data_object_id, data_object)

    def del_provider_data(self, data_object: dict = None) -> None:
        with self.data_containers_lock:
            super().del_provider_data(data_object)

    def get_all_data_containers(self) -> list[dict]:
        with self.data_containers_lock:
            data_containers = super().get_all_data_containers()

        return data_containers

    def search_data_containers(self, data_request: Filter) -> list[dict]:
        with self.data_containers_lock:
            search_result = super().search_data_containers(data_request)

        return search_result

    def check_new_data_recieved(self) -> int:
        with self.data_containers_lock:
            state = super().check_new_data_recieved()

        return state



from __future__ import annotations
import hashlib
import ecdsa


class ECDSABackend:
    """
    Interface for ECDSA backends.
    """

    def create_key(self) -> int:
        """
        Creates a new key pair. Returns the identifier of the key.
        """
        raise NotImplementedError("All the ECDSA backends should implement this")

    def get_public_key(self, identifier: int) -> dict:
        """
        Returns the public key of the key pair with the identifier.
        It returns as a dict to be oer encoded as PublicVerificationKey in oer.
        """
        raise NotImplementedError("All the ECDSA backends should implement this")

    def sign(self, data: bytes, identifier: int) -> dict:
        """
        Signs the array of bytes. With the Key that corresponds to the identifier of the private key.
        The signature must be received in IEEE 1609.2 format.
        """
        raise NotImplementedError("All the ECDSA backends should implement this")

    def verify(self, data: bytes, signature: dict, key: int) -> bool:
        """
        Verifies the array of bytes. With the Key that corresponds to the identifier.
        """
        raise NotImplementedError("All the ECDSA backends should implement this")

    def verify_with_pk(self, data: bytes, signature: dict, pk: dict) -> bool:
        """
        Verifies the array of bytes. With the public key.
        Both key and signature come in IEEE 1609.2 format.

        Parameters
        ----------
        data : bytes
            Data to verify.
        signature : dict
            Signature to verify. In IEEE 1609.2 format (Signature).
        pk : dict
            Public key to verify. In IEEE 1609.2 format (PublicVerificationKey).

        Returns
        -------
        bool
            True if the signature is valid, False otherwise.
        """
        raise NotImplementedError("All the ECDSA backends should implement this")


class PythonECDSABackend(ECDSABackend):
    """
    Python ECDSA backend.

    """

    def __init__(self) -> None:
        """
        Initialize the Python ECDSA backend.
        """
        self.keys: dict[int, ecdsa.keys.SigningKey] = {}

    def create_key(self) -> int:
        """
        Creates a new key pair. Returns the identifier of the key.

        Returns
        -------
        int
            Identifier of the key.
        """
        key: ecdsa.keys.SigningKey = ecdsa.SigningKey.generate(curve=ecdsa.NIST256p)
        identifier = len(self.keys)
        self.keys[identifier] = key
        return identifier

    def get_public_key(self, identifier: int) -> dict:
        """
        Returns the public key of the key pair with the identifier.
        It returns as a dict to be oer encoded as PublicVerificationKey in oer.

        Parameters
        ----------
        identifier : int
            Identifier of the key.

        Returns
        -------
        dict
            Public key in IEEE 1609.2 format.
        """
        key = self.keys[identifier]
        return (
            "ecdsaNistP256",
            (
                "uncompressedP256",
                {
                    "x": key.verifying_key.pubkey.point.x().to_bytes(
                        32, byteorder="big"
                    ),
                    "y": key.verifying_key.pubkey.point.y().to_bytes(
                        32, byteorder="big"
                    ),
                },
            ),
        )

    def sign(self, data: bytes, identifier: int) -> dict:
        """
        Signs the array of bytes. With the Key that corresponds to the identifier of the private key.

        Parameters
        ----------
        data : bytes
            Data to sign.
        identifier : int
            Identifier of the private key.

        Returns
        -------
        dict
            Signature in IEEE 1609.2 format.
        """
        key: ecdsa.keys.SigningKey = self.keys[identifier]
        signature = key.sign(data=data, hashfunc=hashlib.sha256)
        r, s = ecdsa.util.sigdecode_string(signature, ecdsa.NIST256p.order)
        r = r.to_bytes(32, byteorder="big")
        s = s.to_bytes(32, byteorder="big")
        return ("ecdsaNistP256Signature", {"rSig": ("x-only", r), "sSig": s})

    def verify(self, data: bytes, signature: dict, key: int) -> bool:
        """
        Verifies the array of bytes. With the Key that corresponds to the identifier.

        Parameters
        ----------
        data : bytes
            Data to verify.
        signature : dict
            Signature to verify. In IEEE 1609.2 format (Signature).

        Returns
        -------
        bool
            True if the signature is valid, False otherwise.
        """
        for key_, value in self.keys.items():
            if key_ == key:
                vk: ecdsa.keys.VerifyingKey = value.verifying_key
                r = int.from_bytes(signature[1]["rSig"][1], byteorder="big")
                s = int.from_bytes(signature[1]["sSig"], byteorder="big")
                signature = ecdsa.util.sigencode_string(r, s, ecdsa.NIST256p.order)
                return vk.verify(
                    signature=signature, data=data, hashfunc=hashlib.sha256
                )
        return False

    def verify_with_pk(self, data: bytes, signature: dict, pk: dict) -> bool:
        """
        Verifies the array of bytes. With the public key.
        Both key and signature come in IEEE 1609.2 format.

        Parameters
        ----------
        data : bytes
            Data to verify.
        signature : dict
            Signature to verify. In IEEE 1609.2 format (Signature).
        pk : dict
            Public key to verify. In IEEE 1609.2 format (PublicVerificationKey).

        Returns
        -------
        bool
            True if the signature is valid, False otherwise.
        """
        if (
            signature[0] == "ecdsaNistP256Signature"
            and signature[1]["rSig"][0] == "x-only"
        ):
            r = int.from_bytes(signature[1]["rSig"][1], byteorder="big")
            s = int.from_bytes(signature[1]["sSig"], byteorder="big")
            dec_signature = ecdsa.util.sigencode_string(r, s, ecdsa.NIST256p.order)
        else:
            raise ValueError("Signature format not supported")
        if pk[0] == "ecdsaNistP256" and pk[1][0] == "uncompressedP256":
            x = int.from_bytes(pk[1][1]["x"], byteorder="big")
            y = int.from_bytes(pk[1][1]["y"], byteorder="big")
            point = ecdsa.ellipticcurve.Point(
                ecdsa.NIST256p.curve, x, y, ecdsa.NIST256p.order
            )
            vk: ecdsa.keys.VerifyingKey = ecdsa.VerifyingKey.from_public_point(
                point, curve=ecdsa.NIST256p
            )
            try:
                return vk.verify(
                    signature=dec_signature, data=data, hashfunc=hashlib.sha256
                )
            except ecdsa.keys.BadSignatureError:
                return False
        else:
            raise ValueError("Public key format not supported")



from .service_access_point import (
    CommonNH,
    HeaderType,
    HeaderSubType,
    TrafficClass,
    GNDataRequest,
    TopoBroadcastHST,
    GeoBroadcastHST,
    GeoAnycastHST,
    LocationServiceHST,
)
from .exceptions import DecodeError


class CommonHeader:
    """
    Common Header class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.7

    Attributes
    ----------
    nh : CommonNH
        (2 bit unsigned integer) Next Header.
    reserved : int
        (4 bit unsigned integer) Reserved.
    ht : HeaderType
        (4 bit unsigned integer) Header Type (HT) Identifies the Type of Geonetworking Header.
    hst : Enum
        (4 bit unsigned integer) Header Subtype (HST) Identifies the Subtype of Geonetworking Header.
    tc : TrafficClass
        (8 bit unsigned integer) Traffic class that represents Facility-layer requirements on packet transport.
    flags : int
        (8 bit unsigned integer) Flags. Bit 0 Indicates whether the ITS-S is mobile or stationary
        (GN protocol constant itsGnIsMobile) Bit 1 to 7 Reserved.
    pl : int
        (16 bit unsigned integer) Payload Length. Indicates the length of the payload in octets.
    mhl : int
        (8 bit unsigned integer) Maximum Hop Limit. Indicates the maximum number of hops that
        the packet is allowed to traverse.
    reserved : int
        (8 bit unsigned integer) Reserved. Always set to zero.
    """

    def __init__(self) -> None:
        self.nh = CommonNH.ANY
        self.reserved = 0
        self.ht = HeaderType.ANY
        self.hst = HeaderSubType.UNSPECIFIED
        self.tc = TrafficClass()
        self.flags = 0
        self.pl = 0
        self.mhl = 0
        self.reserved = 0

    def initialize_with_request(self, request: GNDataRequest) -> None:
        """
        Initializes the Common Header with a GNDataRequest.

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to use.
        """
        self.nh = request.upper_protocol_entity
        self.ht = request.packet_transport_type.header_type
        self.hst = request.packet_transport_type.header_subtype
        self.tc = request.traffic_class
        self.pl = request.length
        if self.ht == HeaderType.TSB and self.hst == TopoBroadcastHST.SINGLE_HOP:
            self.mhl = 1
        else:
            # TODO: Set the maximum hop limit on other cases than SHB As specified in: Section 10.3.4 Table 20
            self.mhl = 1

    def encode_to_int(self) -> int:
        """
        Encodes the Common Header to an integer.

        Returns
        -------
        int :
            Encoded Common Header.  8 bytes.
        """
        return (
            (self.nh.value << (4 + 8 * 7))
            | (self.reserved << (8 * 7))
            | (self.ht.value << (4 + 8 * 6))
            | (self.hst.value << 8 * 6)
            | (self.tc.encode_to_int() << 8 * 5)
            | self.flags << 8 * 4
            | self.pl << 8 * 2
            | self.mhl << 8
            | self.reserved
        )

    def encode_to_bytes(self) -> bytes:
        """
        Encodes the Common Header to bytes.

        Returns
        -------
        bytes :
            Encoded Common Header. 4 bytes.
        """
        return self.encode_to_int().to_bytes(8, "big")

    def decode_from_int(self, header: int) -> None:
        """
        Decodes an integer to a Common Header.

        Parameters
        ----------
        header : int
            Encoded Common Header. 4 bytes.
        """
        self.nh = CommonNH((header >> (4 + 8 * 7)) & 15)
        self.ht = HeaderType((header >> (4 + 8 * 6)) & 15)
        if self.ht == HeaderType.GEOBROADCAST:
            self.hst = GeoBroadcastHST((header >> (8 * 6)) & 15)
        elif self.ht == HeaderType.TSB:
            self.hst = TopoBroadcastHST((header >> (8 * 6)) & 15)
        elif self.ht == HeaderType.GEOANYCAST:
            self.hst = GeoAnycastHST((header >> (8 * 6)) & 15)
        elif self.ht == HeaderType.LS:
            self.hst = LocationServiceHST((header >> (8 * 6)) & 15)
        else:
            self.hst = HeaderSubType((header >> (8 * 6)) & 15)
        self.tc = TrafficClass()
        self.tc.decode_from_int((header >> 8 * 5) & 255)
        self.flags = (header >> 8 * 4) & 128
        self.pl = (header >> 8 * 2) & 65535
        self.mhl = (header >> 8) & 255

    def decode_from_bytes(self, header: bytes) -> None:
        """
        Decodes bytes to a Common Header.

        Parameters
        ----------
        header : bytes
            Encoded Common Header. 8 bytes.
        """
        if len(header) < 8:
            raise DecodeError("Common Header must be 8 bytes long")
        self.decode_from_int(int.from_bytes(header[0:8], "big"))

    def __eq__(self, __value: object) -> bool:
        """
        Checks if two Common Headers are equal.

        Parameters
        ----------
        __value : object
            Object to compare with.
        """
        if not isinstance(__value, CommonHeader):
            return NotImplemented
        return (
            self.nh == __value.nh
            and self.reserved == __value.reserved
            and self.ht == __value.ht
            and self.hst == __value.hst
            and self.tc == __value.tc
            and self.flags == __value.flags
            and self.pl == __value.pl
            and self.mhl == __value.mhl
            and self.reserved == __value.reserved
        )



import os

from prometheus_client import Histogram, Gauge, CollectorRegistry, start_http_server

PROMETHEUS_PORT = int(os.getenv("PROMETHEUS_CLIENT_PORT", 8000))


class PrometheusClientPull:
    def __init__(self) -> None:
        self.registry = CollectorRegistry()

        self.latency = Histogram("latency", "Average latency of recieved V2X packets in ms", registry=self.registry)
        self.v2x_bandwidth = Histogram(
            "v2x_bandwidth", "Bandwidth used to send V2X packets in Kbps", registry=self.registry
        )
        self.ldm_size = Histogram("ldm_size", "Size of Local Dynamic Maps in Bytes", registry=self.registry)
        self.ldm_map = Gauge(
            "ldm_map",
            "Vehicle Geolocation Data",
            ["station_id", "station_type", "detected_by", "latitude", "longitude"],
        )

        start_http_server(PROMETHEUS_PORT, registry=self.registry)

    def send_ldm_map(self, station_id: str, station_type: str, detected_by: str, latitude: str, longitude: str) -> None:
        """
        Function to expose LDM Map data. It must be called once per LDM Data Object received.

        Parameters
        ----------
        station_id: str
            The ID of the station
        station_type: str
            The type of the station
        detected_by: str
            The device that detected the station
        latitude: str
            The latitude of the station
        longitude: str
            The longitude of the station

        Returns
        ----------
        None
        """
        self.ldm_map.labels(
            station_id=station_id,
            station_type=station_type,
            detected_by=detected_by,
            latitude=str(latitude),
            longitude=str(longitude),
        ).set(1)

    def send_latency(self, value: float) -> None:
        """
        Function to send the latency to the Prometheus Gateway.

        Parameters
        ----------
        value: float
            The value to send to the Prometheus Gateway

        Returns
        -------
        None
        """
        self.latency.observe(value)

    def send_v2x_bandwidth(self, value: float) -> None:
        """
        Function to send the V2X bandwidth to the Prometheus Gateway.

        Parameters
        ----------
        value: float
            The value to send to the Prometheus Gateway

        Returns
        -------
        None
        """
        self.v2x_bandwidth.observe(value)

    def send_ldm_size(self, value: float) -> None:
        """
        Function to send the LDM size to the Prometheus Gateway.

        Parameters
        ----------
        value: float
            The value to send to the Prometheus Gateway

        Returns
        -------
        None
        """
        self.ldm_size.observe(value)



from __future__ import annotations

from .database import DataBase
from .ldm_classes import Filter, RequestDataObjectsReq, Utils
from .ldm_constants import OPERATOR_MAPPING


class DictionaryDataBase(DataBase):
    """
    Class inherting from DataBase.

    This class is used to store data in a dictionary. Its a simple implemenation that can be used for testing or when
    code runs in environments where creating a database (file) is not possible.
    """

    def __init__(self, database_name: str = None, database_path: str = None):
        """
        Parameters
        ----------
        path : str
            Path to the database file.
        """
        self.database = {}

    def delete(self, database_name: str = None) -> bool:
        """
        Delete database with a given name, returns a boolean stating if deletion has been succesful.

        Parameters
        ----------
        database_name : str
            Name of the database to be deleted.
        """
        self.database = {}
        return True

    def search(self, data_request: RequestDataObjectsReq) -> list:
        """
        Search for data with a Filter (from ETSI ETSI EN 302 895 V1.1.1 (2014-09).

        Parameters
        ----------
        query : Filter
            Filter to be used for the search.
        """

        def get_nested(data: dict, path: str) -> int:
            data = data["dataObject"]
            keys = path.split(".")

            for key in keys:
                data = data[key]
            return data

        def create_query_search(query_with_attribute, operator, ref_value):
            if operator in OPERATOR_MAPPING:
                return OPERATOR_MAPPING[operator](query_with_attribute, ref_value)
            raise ValueError(f"Invalid operator: {operator}")

        def filter_data(data_filter: Filter, database: list[dict]) -> str:
            list_of_data = []
            if data_filter.filter_statement_2 is not None:
                if str(data_filter.logical_operator) == "and":
                    for data in database:
                        if create_query_search(
                            get_nested(data, data_filter.filter_statement_1.attribute),
                            str(data_filter.filter_statement_1.operator),
                            data_filter.filter_statement_1.ref_value,
                        ) & create_query_search(
                            get_nested(data, data_filter.filter_statement_2.attribute),
                            str(data_filter.filter_statement_2.operator),
                            data_filter.filter_statement_2.ref_value,
                        ):
                            list_of_data.append(data)
                else:
                    for data in database:
                        if create_query_search(
                            get_nested(data, data_filter.filter_statement_1.attribute),
                            str(data_filter.filter_statement_1.operator),
                            data_filter.filter_statement_1.ref_value,
                        ) | create_query_search(
                            get_nested(data, data_filter.filter_statement_2.attribute),
                            str(data_filter.filter_statement_2.operator),
                            data_filter.filter_statement_2.ref_value,
                        ):
                            list_of_data.append(data)
            else:
                for data in database:
                    if create_query_search(
                        get_nested(data, data_filter.filter_statement_1.attribute),
                        str(data_filter.filter_statement_1.operator),
                        data_filter.filter_statement_1.ref_value,
                    ):
                        list_of_data.append(data)
            return list_of_data

        if data_request.filter is None:
            return RequestDataObjectsReq.filter_out_by_data_object_type(self.all(), data_request.data_object_type)
        try:
            return filter_data(
                data_request.filter,
                RequestDataObjectsReq.filter_out_by_data_object_type(self.all(), data_request.data_object_type),
            )
        except KeyError as e:
            print(f"[ListDatabase] KeyError searching data: {str(e)}")
            return []
        except TypeError as e:
            print(f"[ListDatabase] TypeError searching data: {str(e)}")
            return []

    def insert(self, data: dict) -> int:
        """
        Insert data into the database, returns the index of the inserted data.

        Parameters
        ----------
        data : dict
            Data to be inserted into the database.
        """

        def find_smallest_missing_number(lst):
            lst = set(lst)  # Convert the list to a set for faster lookups
            smallest_number = 0

            while smallest_number in lst:
                smallest_number += 1

            return smallest_number

        index = find_smallest_missing_number(self.database.keys())
        self.database.update({index: data})

        return index

    def get(self, index: int) -> list:
        """
        Get data from the database with a given index.

        Parameters
        ----------
        index : int
            Index of the data to be retrieved.
        """
        try:
            return self.database[index]
        except KeyError:
            return None

    def update(self, data: dict, index: int) -> bool:
        """
        Update data in the database with a given index, returns a boolean stating if update has been succesful.

        Parameters
        ----------
        data : dict
            Data to be updated in the database.
        index : int
            Index of the data to be updated.
        """
        self.database[index] = data
        return True

    def remove(self, data_object: dict = None) -> bool:
        """
        Remove data from the database with a given index, returns a boolean stating if removal has been succesful.

        Parameters
        ----------
        index : int
            Index of the data to be removed.
        """
        list_index = list(self.database.values()).index(data_object)
        dict_index = list(self.database.keys())[list_index]
        self.database.pop(dict_index)
        return True

    def all(self) -> list:
        """
        Get all data from the database.

        Returns
        -------
        list
            All data from the database.
        """
        return list(self.database.values())

    def exists(self, field_name: str, data_object_id: int = None) -> bool:
        """
        Check if specific field exists

        Parameters
        ----------
        field_name : str
            Name of field to be checked
        Returns
        -------
        bool
            Indicates whether field exists
        """
        if field_name == "dataObjectID":
            return data_object_id in self.database
        for data in self.database.values():
            if Utils.check_field(data, field_name):
                return True
        return False



from __future__ import annotations
from .ldm_maintenance import LDMMaintenance

from .ldm_classes import AddDataProviderReq


class LDMMaintenanceReactive(LDMMaintenance):
    """
    Class inheritence from class specified in the ETSI ETSI EN 302 895 V1.1.1 (2014-09).

    In this implementation a reactive apporach is taken to the maintenance of the LDM. This means that the LDM will
    delete data that is not within the area of maintenance only when new data is added. This is done by overriding the
    add_provider_data method.

    """

    def add_provider_data(self, data: AddDataProviderReq) -> int:
        index = super().add_provider_data(data)
        self.logging.debug("Adding provider data; %s", data.data_object)
        self.collect_trash()
        return index



from __future__ import annotations
from collections.abc import Callable
import time
import json


from .ldm_classes import (
    Filter,
    RequestDataObjectsResp,
    RequestedDataObjectsResult,
    RequestDataObjectsReq,
    OrderTuple,
    TimestampIts,
    Utils,
)
from .ldm_constants import (
    DATA_OBJECT_TYPE_ID,
)
from .ldm_maintenance import (
    LDMMaintenance,
)


class LDMService:
    """
    Class specified in the ETSI ETSI EN 302 895 V1.1.1 (2014-09). Section 5.3.
    The Local Dynamic Map (LDM) Service component is responsible for providing functionalities to authorized LDM data
    Providers and Consumers for LDM data manipulation
    (such as adding new data, modifying existing data, delete existing data), direct access to data (query data)
    and a publish/subscribe mechanism for data access by LDM Data Consumers

    Attributes
    ----------
    TODO: Add attributes

    """

    def __init__(self, ldm_maintenance: LDMMaintenance) -> None:
        self.ldm_maintenance = ldm_maintenance

        self.data_provider_its_aid: list[int] = []
        self.data_consumer_its_aid: list[int] = []

        self.subscriptions: list[dict] = []

    def subscriptions_service(self) -> None:
        """
        Method that attends subscriptions.

        The method is meant to be run in a seperate thread.
        """
        while True:
            self.attend_subscriptions()

    def attend_subscriptions(self) -> None:
        """
        Method to attend subscriptions as specified in the ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4.
        """
        for subscription in self.subscriptions:
            search_result = self.search_data(subscription)
            if not search_result:
                continue

            valid_search_result = self.filter_data_object_type(
                search_result, subscription["data_object_type"]
            )
            if subscription["multiplicity"] is not None and subscription[
                "multiplicity"
            ] > len(valid_search_result):
                continue

            if subscription["order"] is not None:
                valid_search_result = self.order_search_results(
                    valid_search_result, subscription["order"]
                )

            self.process_notifications(subscription, valid_search_result)

            if subscription["applicationId"] not in self.data_consumer_its_aid:
                self.pop_subscription(subscription["doc_id"])

    def search_data(self, subscription: dict) -> list[dict]:
        """
        Method to search data by using the filter in the subscription.

        Parameters
        ----------
        subscription : dict
            Subscription information (in dictionary format).
        Returns
        -------
        list[dict]
            list of data objects that match the filter.
        """
        data_request = RequestDataObjectsReq(
            subscription["applicationId"],
            subscription["data_object_type"],
            subscription["priority"],
            subscription["order"],
            subscription["filter"],
        )
        return self.ldm_maintenance.data_containers.search(data_request)

    def filter_data_object_type(
        self, search_result: list[dict], allowed_types: list[str]
    ) -> list[dict]:
        """
        Method to filter data objects by type.

        Parameters
        ----------
        search_result : list[dict]
            list of data objects that match the filter.
        allowed_types : list[str]
            list of allowed data object types
            (as specified in the ETSI TS 102 894-2 V2.2.1 (2023-10), i.e. CAM, DENM,...).

        Returns
        -------
        list[dict]
            list of data objects that match the filter and are of the allowed types.
        """
        return [
            result
            for result in search_result
            if self.get_object_type_from_data_object(result["dataObject"])
            in allowed_types
        ]

    def process_notifications(
        self, subscription: dict, valid_search_result: list[dict]
    ) -> None:
        """
        Method to process notifications for a subscription. It checks the notification interval and the last time
        it had been checked. If the notification interval has passed since the last check, it will send a notification
        to the callback function (listed in the subscription information dictionary database).

        Parameters
        ----------
        subscription : dict
            Subscription information (in dictionary format).
        valid_search_result : list[dict]
            list of data objects that match the filter and are of the allowed types.

        Returns
        -------
        None
        """
        current_time = time.time()
        if (
            subscription["last_checked"] + subscription["notification_interval"]
            > current_time
        ):
            return

        subscription["last_checked"] = current_time
        subscription["callback"](
            RequestDataObjectsResp(
                application_id=subscription["applicationId"],
                data_objects=valid_search_result,
                result=RequestedDataObjectsResult(result=0),
            )
        )

    def pop_subscription(self, index: int) -> None:
        """
        Method that removes (pops) a subscription from the subscriptions list and updates the new index values
        in the "doc_id" field.

        Parameters
        ----------
        index : int
        """
        if index < 0 or index >= len(self.subscriptions):
            raise IndexError("Index out of range")

        self.subscriptions.pop(index)

        for i, _ in enumerate(self.subscriptions):
            self.subscriptions[i]["doc_id"] = i

    def find_key_paths_in_list(self, target_key: str, search_result: list) -> list[str]:
        """
        Static method to find the path of a key in a list

        Parameters
        ----------
        target_key : str
        search_result : list
        """
        key_paths = []
        for result in search_result:
            key_paths.append(self.find_key_path(target_key, result))
        return key_paths

    def find_key_path(self, target_key: str, dictionary: dict, path: str = None) -> str:
        """
        Static method to find the path of a key in a dictionary.

        Parameters
        ----------
        target_key : str
        dictionary : dict
        path : list
        """
        if path is None:
            path = []
        for key, value in dictionary.items():
            if key == target_key:
                path.append(key)
                return ".".join(path)  # Return the formatted path
            if isinstance(value, dict):
                sub_path = self.find_key_path(target_key, value, path + [key])
                if sub_path:
                    return sub_path
        return None

    def order_search_results(
        self, search_results: list[dict], orders: list[OrderTuple]
    ) -> list[list[dict]]:
        """
        Method to order search results.

        Parameters
        ----------
        search_results : list[dict]
        orders : list[OrderTuple]
        """
        results = []
        for order in orders:
            tuple_list = [
                (
                    index,
                    Utils.get_nested(
                        search_result,
                        Utils.find_attribute(order.attribute, search_result),
                    ),
                )
                for index, search_result in enumerate(search_results)
            ]
            if str(order.ordering_direction) == "ascending":
                sorted_tuple_list = sorted(
                    tuple_list, key=lambda x: x[1], reverse=False
                )
            else:
                sorted_tuple_list = sorted(tuple_list, key=lambda x: x[1], reverse=True)

            result = [search_results[tuple[0]] for tuple in sorted_tuple_list]
            results.append(result)
        return results

    def add_provider_data(self, data: dict) -> int:
        """
        Method created in order to add provider data into the data containers.

        Parameters
        ----------
        data : dict
        """
        return self.ldm_maintenance.add_provider_data(data)

    def add_data_provider_its_aid(self, its_aid: int) -> None:
        """
        Method created in order to add the providers ITS_AID into the list of data consumers.

        Parameters
        ----------
        its_aid : int
        """
        self.data_provider_its_aid.append(its_aid)

    def update_provider_data(self, data_object_id: int, data_object: dict) -> None:
        """
        Method used to update provider data.

        Parameters
        ----------
        data_object_id : int
        data_object : dict
        """
        self.ldm_maintenance.update_provider_data(data_object_id, data_object)

    def get_provider_data(self) -> list:
        """
        Method to get provider data from the data containers.

        Parameters
        ----------
        None
        """
        return self.ldm_maintenance.data_containers

    def get_data_provider_its_aid(self) -> list[int]:
        """
        Method to get all providers ITS_AID values.

        Parameters
        ----------
        None
        """
        return self.data_provider_its_aid

    def del_provider_data(self, provider_data_id: int) -> None:
        """
        Method to delete all provider data with doc_id.

        Parameters
        ----------
        provider_data_id : int
        """
        self.data_provider_its_aid.remove(provider_data_id)

    def del_data_provider_its_aid(self, its_aid: int) -> None:
        """
        Method to delete provider ITS_AID from the list of data providers.

        Parameters
        ----------
        its_aid : int
        """
        self.data_provider_its_aid.remove(its_aid)

    def query(self, data_request: RequestDataObjectsReq) -> list[list[dict]]:
        """
        Method to query the data containers using the RequestDataObjectsReq object.

        Parameters
        ----------
        data_request : RequestDataObjectsReq

        Returns
        -------
        list[list[dict]]
            list of list of data objects. Each list of data objects is ordered according to the order specified in the data_request object.
        """
        try:
            if data_request.filter is None:
                search_result = self.ldm_maintenance.get_all_data_containers()
            else:
                search_result = self.ldm_maintenance.search_data_containers(
                    data_request
                )
        except (KeyError, json.decoder.JSONDecodeError) as e:
            print(f"Error querying data container: {str(e)}")
            return [[]]

        # If it does then see if it needs ordering and order it
        if data_request.order is not None:
            search_result = self.order_search_results(search_result, data_request.order)
        else:
            search_result = [search_result]
        return search_result

    def get_object_type_from_data_object(self, data_object: dict) -> str:
        """
        Method to get object type from data object.

        Parameters
        ----------
        data_object : dict
        """
        for data_object_type_str in data_object.keys():
            if data_object_type_str in DATA_OBJECT_TYPE_ID.values():
                return list(DATA_OBJECT_TYPE_ID.keys())[
                    list(DATA_OBJECT_TYPE_ID.values()).index(data_object_type_str)
                ]
        return None

    def store_new_subscription_petition(
        self,
        application_id: int,
        data_object_type: list[int],
        priority: int,
        filter: Filter,
        notification_interval: TimestampIts,
        multiplicity: int,
        order: list[OrderTuple],
        callback: Callable[[RequestDataObjectsResp], None],
    ) -> int:
        # pylint: disable=too-many-arguments
        """
        Method as standarized in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4.
        Method is used to store a new subscription petition.

        Parameters
        ----------
        application_id : int
        data_object_type : int
        priority : int
        filter : Filter
        notification_interval : int
        multiplicity : int
        order : int
        callback : function
        """
        return self.subscriptions.append(
            {
                "applicationId": application_id,
                "data_object_type": data_object_type,
                "priority": priority,
                "filter": filter,
                "notification_interval": notification_interval,
                "multiplicity": multiplicity,
                "order": order,
                "callback": callback,
                "last_checked": time.time(),
                "doc_id": len(self.subscriptions),
            }
        )

    def add_data_consumer_its_aid(self, its_aid: int) -> None:
        """
        Method to add data consumer ITS_AID to the list of data consumers.

        Parameters
        ----------
        its_aid : int
        """
        self.data_consumer_its_aid.append(its_aid)

    def get_data_consumer_its_aid(self) -> list[int]:
        """
        Method to get list of data consumer ITS_AID.

        Parameters
        ----------
        None
        """
        return self.data_consumer_its_aid

    def del_data_consumer_its_aid(self, its_aid: int) -> None:
        """
        Method to delete data consumer ITS_AID from the list of data consumers.

        Parameters
        ----------
        its_aid : int
        """
        self.data_consumer_its_aid.remove(its_aid)

    def get_data_consumer_subscriptions(self) -> list[dict]:
        """
        Method to get data consumer subscriptions from the subscriptions storage.

        Parameters
        ----------
        None
        """
        return self.subscriptions

    def delete_subscription(self, subscription_id: int) -> bool:
        """
        Method to delete subscriptions from the subscriptions storage.

        Parameters
        ----------
        subscription_id : int
        """
        try:
            self.subscriptions.pop(subscription_id)
            return True
        except IndexError:
            return False



"""
CAM Coder.

This file contains the class for the CAM Coder.
"""
import asn1tools
from .cam_asn1 import CAM_ASN1_DESCRIPTIONS


class CAMCoder:
    """
    Class for encoding/decoding Cooperative Awareness Messages (CAM)

    Attributes
    ----------
    asn_coder : asn1tools.Coder
        ASN.1 coder.
    """

    def __init__(self) -> None:
        """
        Initialize the CAM Coder.
        """
        self.asn_coder = asn1tools.compile_string(CAM_ASN1_DESCRIPTIONS, codec="uper")

    def encode(self, cam: dict) -> bytes:
        """
        Encodes a CAM message.

        Parameters
        ----------
        cam : dict
            CAM message.

        Returns
        -------
        bytes
            Encoded CAM message.
        """
        return self.asn_coder.encode("CAM", cam)

    def decode(self, cam: bytes) -> dict:
        """
        Decodes a CAM message.

        Parameters
        ----------
        cam : bytes
            Encoded CAM message.

        Returns
        -------
        dict
            CAM message.
        """
        return self.asn_coder.decode("CAM", cam)



import time


class TimeService:
    """
    Time Service that provides the current time.

    This class serves as a blueprint in implementations where the time comes
    from other devices or services different from the system time.
    """

    @staticmethod
    def time() -> float:
        """
        Get the current time.

        Returns
        -------
        float
            The current time.
        """
        return time.time()



from .sn_sap import ReportVerify, SNVERIFYRequest, SNVERIFYConfirm
from .sign_service import ECDSABackend
from .security_coder import SecurityCoder
from .certificate_library import CertificateLibrary


class VerifyService:
    """
    Class to verify the signature of a message
    """

    def __init__(
        self,
        backend: ECDSABackend,
        security_coder: SecurityCoder,
        certificate_library: CertificateLibrary,
    ):
        """
        Constructor

        """
        self.backend: ECDSABackend = backend
        self.security_coder: SecurityCoder = security_coder
        self.certificate_library: CertificateLibrary = certificate_library

    def verify(self, request: SNVERIFYRequest) -> SNVERIFYConfirm:
        """
        Verify the signature of a message
        """
        verify_confirm = SNVERIFYConfirm()

        sec_header_decoded = self.security_coder.decode_etsi_ts_103097_data_signed(
            request.message
        )
        data = self.security_coder.encode_to_be_signed_data(
            sec_header_decoded["toBeSigned"]
        )
        signer = sec_header_decoded["content"][1]["signer"]
        authorization_ticket = None
        if signer[0] == "certificate":
            authorization_ticket = (
                self.certificate_library.verify_sequence_of_certificates(
                    signer[1], self.security_coder, self.backend
                )
            )
            if not authorization_ticket:
                verify_confirm.report = ReportVerify.INCONSISTENT_CHAIN
                return verify_confirm
        elif signer[0] == "digest":
            authorization_ticket = (
                self.certificate_library.get_authorization_ticket_by_hashedid8(
                    signer[1]
                )
            )
            if not authorization_ticket:
                verify_confirm.report = ReportVerify.INVALID_CERTIFICATE
                return verify_confirm
        else:
            raise Exception("Unknown signer type")
        if (
            authorization_ticket is not None
            and authorization_ticket.verify(authorization_ticket.issuer)
            and authorization_ticket["toBeSigned"]["verifyKeyIndicator"][0]
            == "verificationKey"
        ):
            verification_key = authorization_ticket["toBeSigned"]["verifyKeyIndicator"][
                1
            ]
            verify = self.backend.verify_with_pk(
                data=data,
                signature=sec_header_decoded["signature"],
                pk=verification_key,
            )
            if verify:
                verify_confirm.report = ReportVerify.SUCCESS
                verify_confirm.certificate_id = authorization_ticket.as_hashedid8()
                verify_confirm.its_aid = b""
                verify_confirm.its_aid_length = 0
                verify_confirm.permissions = b""
            else:
                verify_confirm.report = ReportVerify.FALSE_SIGNATURE
        else:
            verify_confirm.report = ReportVerify.INVALID_CERTIFICATE
        return verify_confirm



from .ldm_constants import (
    RESULT_INVALID_ITS_AID,
    RESULT_NOT_AUTHORIZED,
    RESULT_UNABLE_APPLICATION_AUTHENTICATION,
)


class InterfaceLDM2:
    """
    Class specified is described in ETSI EN 302 895 V1.1.1 (2014-09). Section 5.4.2.
    The IF.LDM.2 interface is responsible for the exchange of information with the ITS Security Layer, as described
    in EN 302 665.
    The ITS security layer will exchange information with the LDM across interface IF.LDM.2 in order to revoke the
    authorization of previously authorized iTS LDM Data Provider/Consumer.
    TODO: Finish once Security Layer is implemented

    Attributes
    ----------
    None
    """

    def __init__(self) -> None:
        pass


class Authorization:
    """
    Class to handle Authorization requests.

    """

    def __init__(self, its_application_identifider: str, permissions_granted: int):
        self.its_application_identifider = its_application_identifider
        self.permissions_granted = permissions_granted
        self.result = 0

    def check_its_aid(self) -> bool:
        """
        Check if the ITS AID is valid.

        Returns
        -------
        bool
            True if ITS AID is valid, False otherwise.
        """

    def check_permissions(self) -> bool:
        """
        Check permissions granted.

        Returns
        -------
        bool
            True if permissions granted, False otherwise.
        """
        # TODO: Permissions given could be different than permissions requested.

    def check_application_authentication(self) -> bool:
        """
        Check application authentication.

        Returns
        -------
        bool
            True if application authenticated, False otherwise.
        """

    def generate_result(self) -> str:
        """
        Check if the ITS AID is valid, permissions granted and application authenticated.
        If all checks are passed, return RESULT_OK.

        Returns
        -------
        str
            RESULT_OK if all checks are passed, otherwise return the corresponding error.
        """
        if not self.check_its_aid():
            return RESULT_INVALID_ITS_AID
        if not self.check_permissions():
            return RESULT_NOT_AUTHORIZED
        if not self.check_application_authentication():
            return RESULT_UNABLE_APPLICATION_AUTHENTICATION

        raise NotImplementedError("TODO: Finish once Security Layer is implemented")

    def __str__(self) -> str:
        return f"Its_application_identifider: {self.its_application_identifider}, \
            Permissions_granted: {self.permissions_granted}, Result: {self.result}"



from .exceptions import DecodeError
from .position_vector import LongPositionVector
from .service_access_point import GNDataRequest


class GBCExtendedHeader:
    """
    GBC Extended Header class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.8 (Table 14 & Table 36)

    Attributes
    ----------
    sn : int
        Sequence number.
    reserved : int
        Reserved. All bits set to zero.
    so_pv : LongPositionVector
        Source Long Position Vector.
    latitude : int
        Latitude of the center of the area. In 1/10 micro degree.
    longitude : int
        Longitude of the center of the area. In 1/10 micro degree.
    a : int
        Length of the semi-major axis. In meters.
    b : int
        Length of the semi-minor axis. In meters.
    angle : int
        In degrees from North
    reserved2 : int
        Reserved. All bits set to zero.
    """

    def __init__(self) -> None:
        self.sn = 0
        self.reserved = 0
        self.so_pv = LongPositionVector()
        self.latitude = 0
        self.longitude = 0
        self.a = 0
        self.b = 0
        self.angle = 0
        self.reserved2 = 0

    def initialize_with_request(self, request: GNDataRequest) -> None:
        """
        Initialize the GBC Extended Header with a GN Data Request.

        Parameters
        ----------
        request : GNDataRequest
            The GN Data Request.
        """
        self.latitude = request.area.latitude
        self.longitude = request.area.longitude
        self.a = request.area.a
        self.b = request.area.b
        self.angle = request.area.angle

    def encode(self) -> bytes:
        """
        Encode the GBC Extended Header to bytes.

        Returns
        -------
        bytes
            The encoded bytes.
        """
        return (
            self.sn.to_bytes(2, "big")
            + self.reserved.to_bytes(2, "big")
            + self.so_pv.encode()
            + self.latitude.to_bytes(4, "big")
            + self.longitude.to_bytes(4, "big")
            + self.a.to_bytes(2, "big")
            + self.b.to_bytes(2, "big")
            + self.angle.to_bytes(2, "big")
            + self.reserved2.to_bytes(2, "big")
        )

    def decode(self, header: bytes) -> None:
        """
        Decode the GBC Extended Header from bytes.

        Parameters
        ----------
        header : bytes
            The header bytes.

        Raises
        ------
        DecodeError
            If the header is not 44 bytes long.
        """
        if len(header) < 44:
            raise DecodeError("GBC Extended Header must be 44 bytes long")
        self.sn = int.from_bytes(header[0:2], "big")
        self.reserved = int.from_bytes(header[2:4], "big")
        self.so_pv.decode(header[4:28])
        self.latitude = int.from_bytes(header[28:32], "big")
        self.longitude = int.from_bytes(header[32:36], "big")
        self.a = int.from_bytes(header[36:38], "big")
        self.b = int.from_bytes(header[38:40], "big")
        self.angle = int.from_bytes(header[40:42], "big")
        self.reserved2 = int.from_bytes(header[42:44], "big")

    def __str__(self) -> str:
        return (
            "GBC Extended Header"
            + "\n"
            + "Sequence number: "
            + str(self.sn)
            + "\n"
            + "Reserved: "
            + str(self.reserved)
            + "\n"
            + "Source Long Position Vector: \n"
            + str(self.so_pv)
            + "\n"
            + "Latitude: "
            + str(self.latitude)
            + "\n"
            + "Longitude: "
            + str(self.longitude)
            + "\n"
            + "Length of the semi-major axis: "
            + str(self.a)
            + "\n"
            + "Length of the semi-minor axis: "
            + str(self.b)
            + "\n"
            + "Angle: "
            + str(self.angle)
            + "\n"
            + "Reserved: "
            + str(self.reserved2)
            + "\n"
        )



from __future__ import annotations

import logging

from dateutil import parser

from ..local_dynamic_map.ldm_classes import Utils

from .vam_ldm_adaptation import VRUBasicServiceLDM

from .vam_coder import VAMCoder
from ...btp.router import Router as BTPRouter
from ...btp.service_access_point import (
    BTPDataRequest,
    CommonNH,
    PacketTransportType,
    CommunicationProfile,
    TrafficClass,
)
from ..ca_basic_service.cam_transmission_management import (
    GenerationDeltaTime,
    CooperativeAwarenessMessage,
)
from . import vam_constants

# from ..local_dynamic_map import LocalDynamicMap


class DeviceDataProvider:
    """
    Class that stores the vehicle data.

    Attributes
    ----------
    station_id : int
        Station ID as specified in ETSI TS 103 300-3 V2.2.1 (2023-02).
    station_type : int
        Station Type as specified in ETSI TS 103 300-3 V2.2.1 (2023-02).
    heading : dict
       Heading as specified in ETSI TS 103 300-3 V2.2.1 (2023-02).
    speed : dict
        Speed as specified in ETSI TS 103 300-3 V2.2.1 (2023-02).
    longitudinalAcceleration : dict
        Longitudinal Acceleration as specified in ETSI TS 103 300-3 V2.2.1 (2023-02).
    """

    def __init__(self) -> None:
        self.station_id = 0
        self.station_type = 0
        self.heading = {"value": 3601, "confidence": 127}
        self.speed = {"speedValue": 16383, "speedConfidence": 127}
        self.longitudinal_acceleration = {
            "longitudinalAccelerationValue": 161,
            "longitudinalAccelerationConfidence": 102,
        }


class PathPoint:
    """
    Class that stores the path point data.

    Attributes
    ----------
    pathPosition : tuple
        The pathPosition DF shall comprise the latitude, longitude and altitude of the VRU.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
    pathDeltaTime : int
        The pathDeltaTime DF shall comprise the time difference between the current time and
        the time when the VRU was at the pathPosition.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
    """

    def __init__(self, latitude, longitude, altitude, time=None):
        self.path_position = (latitude, longitude, altitude)
        self.path_delta_time = time


class PathHistory:
    """
    Class that stores the path history data.

    Attributes
    ----------
    pathPoints : list
        The pathPoints DF shall comprise the VRU's recent movement over the last 40 pathPoints.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
    """

    def __init__(self, path_points: list[PathPoint] = None):
        self.path_points = path_points

    def append(self, path_point):
        """
        Appends each point to the pathPoints list. With a maximum value of 40
        (as specified by ETSI TS 103 300-3 V2.2.1 (2023-02))

        Attributes
        ----------
        path_point : PathPoint

        """
        try:
            if len(self.path_points) >= 40:
                self.path_points.pop(0)
            self.path_points.append(path_point)
        except TypeError:
            self.path_points = [path_point]

    def generate_path_history_dict(self):
        """
        Generates the path history JSON data.

        Attributes
        ----------
        path_points : list
            The pathPoints DF shall comprise the VRU's recent movement over the last 40 pathPoints.
            All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
        """
        path_points = []
        for path_point in self.path_points:
            path_points.append(
                {
                    "pathPosition": {
                        "latitude": path_point.path_position[0],
                        "longitude": path_point.path_position[1],
                        "altitude": path_point.path_position[2],
                    },
                    "pathDeltaTime": path_point.path_delta_time,
                }
            )
        return path_points


class PathPointPredicted:
    """
    Class that stores the path point predicted data.

    Attributes
    ----------
    deltaLatitude : int
        The deltaLatitude DF shall comprise an offset latitude with regards to a
        pre-defined reference position.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
    deltaLongitude : int
        The deltaLongitude DF shall comprise an offset longitude with regards to a
        pre-defined reference position.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
    pathDeltaTime : int
        The pathDeltaTime DF shall comprise the  travel time separated from the
        waypoint to the predefined reference position.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).

    Attributes: horizontalPositionConfidence (OPTIONAL), deltaAltitude (UNAVAILABLE),
                altitudeConfidence (UNAVAILABLE) are not currently implemented.
    """

    def __init__(self, delta_latitude, delta_longitude, path_delta_time):
        self.delta_latitude = delta_latitude
        self.delta_longitude = delta_longitude
        self.path_delta_time = path_delta_time

    def __dict__(self) -> dict:
        return {
            "deltaLatitude": self.delta_latitude,
            "deltaLongitude": self.delta_longitude,
            "pathDeltaTime": self.path_delta_time,
        }


class PathPrediction:
    """
    Class that stores the path prediction data.

    Attributes
    ----------
    pathPointPredicted : list
        The pathPoints DF shall comprise the VRU's recent movement over the last 40 pathPoints.
        All values as specified by ETSI TS 103 300-3 V2.2.1 (2023-02).
    """

    def __init__(self, path_point_predicted: list[PathPointPredicted] = None):
        self.path_point_predicted = path_point_predicted

    def append(self, path_point: PathPointPredicted):
        """
        Function to append data into the PathPrediction class. It will append data until the
        maximum length of 15 is achieved. Then it will act as a FIFO memory.

        Parameters
        -----------
        path_point: PathPointPredicted
            The path point predicted

        Returns
        --------
        None
        """
        try:
            if len(self.path_point_predicted) >= 15:
                self.path_point_predicted.pop(0)
            self.path_point_predicted.append(path_point)
        except TypeError:
            self.path_point_predicted = [path_point]

    def generate_path_prediction_dict(self):
        """
        Generates a json object from the PathPrediction class.

        Returns
        -------
        dict
            A dictionary containing the PathPrediction class attributes.
        """
        return {
            "pathPointPredicted": [
                path_point.__dict__() for path_point in self.path_point_predicted
            ]
        }


class MotionPredictionContainer:
    """
    TODO: Implement the Motion Prediction Container using the Local Dynamic Map (LDM).

    As specified by ETSI TS 103 300-3 V2.2.1 (2023-02);

    The VRU Motion Prediction Container carries the past and future motion state information of
    the VRU. The VRU Motion Prediction Container of type VruMotionPredictionContainer shall contain
    information about the past locations of the VRU of type PathHistory, predicted future locations
    of the VRU, safe distance indication between VRU and other road users/objects of type
    SequenceOfSafeDistanceIndication, VRU's possible trajectory interception with another
    VRU/object shall be of type SequenceOfTrajectoryInterceptionIndication , the change in the
    acceleration of the VRU shall be of type AccelerationChangeIndication, the heading changes of
    the VRU shall be of HeadingChangeIndication, and changes in the stability of the VRU shall be
    of type StabilityChangeIndication:

    - The pathHistory is of PathHistory type. The PathHistory DF shall comprise the VRU's recent
        movement over past time and/or distance. It consists of up to 40 past path points
        (see ETSI TS 102 894-2 [7]). When a VRU leaves a cluster and wants to transmit its past
        locations in the VAM, the VRU may use the PathHistory DF.
    - The Path Prediction DF is of PathPredicted type and shall define up to 15 future path points,
        confidence values and corresponding time instances of the VRU ITS-S. It contains future
        path information for up to 10 seconds or up to 15 path points, whichever is smaller.
    - The Safe Distance Indication is of type SequenceOfSafeDistanceIndication and provides an
        indication of whether the VRU is at a recommended safe distance laterally,
        longitudinally and vertically from up to 8 other stations in its vicinity. The simultaneous
        comparisons between Lateral Distance (LaD), Longitudinal Distance (LoD) and
        Vertical Distance (VD) and their respective thresholds, Minimum Safe Lateral Distance
        (MSLaD), Minimum Safe Longitudinal Distance (MSLoD), and Minimum Safe Vertical Distance
        (MSVD) as defined inclause 6.5.10.5 of ETSI TS 103 300-2 [1], shall be used for setting the
        safeDistanceIndicator DF. Other ITS-s involved are indicated as subjectStation DE within
        the SafeDistanceIndication DE. The timeToCollision (TTC) DE within the container shall
        reflect the estimated time taken for collision based on the latest onboard sensor
        measurements and VAMs.
    - The trajectoryInterceptionIndication shall contain ego-VRU's possible trajectory interception
        with up to 8 other stations in the vicinity of the ego-VRU. The trajectory interception of
        a VRU is indicated by trajectoryInterceptionIndication DF. The other ITS-S involved are
        designated by StationID DE. The trajectory interception probability and its confidence
        level metrics are indicated by TrajectoryInterceptionProbability and
        TrajectoryInterceptionConfidence DEs.
    - The accelerationChangeIndication shall contain ego-VRU's change of acceleration in the
        future (acceleration or deceleration) for a time period. The DE accelOrDecel shall give the
        choice between acceleration and deceleration. The DE actionDeltaTime shall indicate the
        time duration.
    - The headingChangeIndication shall contain ego-VRU's change of heading in the future
        (left or right) for a time period. The DE directionshall give the choice between heading
        change in left and right directions. The DE actionDeltaTime shall indicate the time
        duration.
    - The stabilityChangeIndication shall contain ego-VRU's change in stability for a time period.
        The lossProbability shall give the probability indication of the stability loss of the
        ego-VRU. It is expressed in the estimated probability of a complete VRU stability loss
        which may lead to a VRU ejection of its VRU vehicle. The loss of stability is projected
        for a time period actionDeltaTime.
    """

    def __init__(
        self,
        path_history: PathHistory,
        path_prediction: PathPrediction,
        safe_distance=None,
        trajectory_interception_indication=None,
        acceleration_change_indication=None,
        heading_change_indication=None,
        stability_change_indication=None,
    ) -> None:
        self.path_history = path_history
        self.path_prediction = path_prediction
        self.safe_distance = safe_distance
        self.trajectory_interception_indication = trajectory_interception_indication
        self.acceleration_change_indication = acceleration_change_indication
        self.heading_change_indication = heading_change_indication
        self.stability_change_indication = stability_change_indication

    def generate_motion_container_message(self) -> dict:
        """
        Function to Generate Motion Containers (as specified in ETSI TS 103 300-3 V2.2.1 (2023-02)).

        Parameters
        ----------
        None
        """
        motion_prediction_container_json = {
            "pathHistory": self.path_history.generate_path_history_dict(),
            "pathPrediction": self.path_prediction.generate_path_prediction_dict(),
            "safeDistance": self.safe_distance,
            "trajectoryInterceptionIndication": self.trajectory_interception_indication,
            "accelerationChangeIndication": self.acceleration_change_indication,
            "headingChangeIndication": self.heading_change_indication,
            "stabilityChangeIndication": self.stability_change_indication,
        }
        return motion_prediction_container_json


class VAMMessage(CooperativeAwarenessMessage):
    """
    VAM Message class.

    Attributes
    ----------
    vam : dict
        All the vam message in dict format as decoded by the vamCoder.

    """

    def __init__(self) -> None:
        super().__init__()
        self.vam = self.generate_white_vam()

    def generate_white_vam(self) -> dict:
        """
        Generate a white vam.
        """
        white_vam = {
            "header": {"protocolVersion": 3, "messageId": 16, "stationId": 0},
            "vam": {
                "generationDeltaTime": 0,
                "vamParameters": {
                    "basicContainer": {
                        # roadSideUnit(15), cyclist(2)
                        "stationType": 15,
                        "referencePosition": {
                            "latitude": 900000001,
                            "longitude": 1800000001,
                            "positionConfidenceEllipse": {
                                "semiMajorAxisLength": 4095,
                                "semiMinorAxisLength": 4095,
                                "semiMajorAxisOrientation": 3601,
                            },
                            "altitude": {
                                "altitudeValue": 800001,
                                "altitudeConfidence": "unavailable",
                            },
                        },
                    },
                    "vruHighFrequencyContainer": {
                        "heading": {"value": 3601, "confidence": 127},
                        "speed": {"speedValue": 16383, "speedConfidence": 127},
                        "longitudinalAcceleration": {
                            "longitudinalAccelerationValue": 161,
                            "longitudinalAccelerationConfidence": 102,
                        },
                    },
                },
            },
        }

        return white_vam

    def fullfill_with_device_data(
        self, device_data_provider: DeviceDataProvider
    ) -> None:
        """
        Fullfill the vam with vehicle data.
        """
        self.vam["header"]["stationId"] = device_data_provider.station_id
        self.vam["vam"]["vamParameters"]["basicContainer"][
            "stationType"
        ] = device_data_provider.station_type
        self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"][
            "heading"
        ] = device_data_provider.heading
        self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"][
            "speed"
        ] = device_data_provider.speed
        self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"][
            "longitudinalAcceleration"
        ] = device_data_provider.longitudinal_acceleration

    def fullfill_gen_delta_time_with_tpv_data(self, tpv: dict) -> None:
        """
        Fullfills the generation delta time with the GPSD TPV data.

        Attributes
        ----------
        tpv : dict
            GPSD TPV data.
        """
        if "time" in tpv:
            self.vam["vam"]["generationDeltaTime"] = int(
                (((parser.parse(tpv["time"]).timestamp()) - 1072911600) * 1000) % 65536
            )

    def create_position_confidence(self, epx: int, epy: int) -> dict:
        """
        Translates the epx and epy TPV values to the position confidence ellipse value.

        Parameters
        ----------
        epx : int
            TPV epx value.
        epy : int
            TPV epy value.

        Returns
        -------
        dict
            Position confidence ellipse value.
        """
        position_confidence_ellipse = {
            "semiMajorAxisLength": int(epx * 100),
            "semiMinorAxisLength": int(epy * 100),
            "semiMajorAxisOrientation": 0,
        }

        return position_confidence_ellipse

    def fullfill_basic_container_with_tpv_data(self, tpv: dict) -> None:
        """
        Fullfills the basic container with the GPSD TPV data.

        Attributes
        ----------
        tpv : dict
            GPSD TPV data.
        """
        if "lat" in tpv.keys():
            self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                "latitude"
            ] = int(tpv["lat"] * 10000000)
        if "lon" in tpv.keys():
            self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                "longitude"
            ] = int(tpv["lon"] * 10000000)
        if "epx" in tpv.keys() and "epy" in tpv.keys():
            self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                "positionConfidenceEllipse"
            ] = self.create_position_confidence(tpv["epx"], tpv["epy"])
        if "altHAE" in tpv.keys():
            alt = int(tpv["altHAE"] * 100)
            if alt < -800000:
                self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                    "altitude"
                ]["altitudeValue"] = -100000
            elif alt > 613000:
                self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                    "altitude"
                ]["altitudeValue"] = 800000
            else:
                self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                    "altitude"
                ]["altitudeValue"] = int(tpv["altHAE"] * 100)
        if "epv" in tpv.keys():
            self.vam["vam"]["vamParameters"]["basicContainer"]["referencePosition"][
                "altitude"
            ]["altitudeConfidence"] = self.create_altitude_confidence(tpv["epv"])

    def fullfill_high_frequency_container_with_tpv_data(self, tpv: dict) -> None:
        """
        Fullfills the high frequency container with the GPSD TPV data.
        Attributes
        ----------
        tpv : dict
            GPSD TPV data.
        """
        if "track" in tpv.keys():
            self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"]["heading"][
                "headingValue"
            ] = int(tpv["track"])
        if "epd" in tpv.keys():
            self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"]["heading"][
                "headingConfidence"
            ] = self.create_heading_confidence(tpv["epd"])
        if "speed" in tpv.keys():
            if int(tpv["speed"] * 100) > 16381:
                self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"]["speed"][
                    "speedValue"
                ] = 16382
            else:
                self.vam["vam"]["vamParameters"]["vruHighFrequencyContainer"]["speed"][
                    "speedValue"
                ] = int(tpv["speed"] * 100)


class VAMTransmissionManagement:
    """
    vam Transmission Management class.
    This sub-function implements the protocol operation of the originating ITS-S, as specified in
    ETSI TS 103 300-3 V2.2.1 clause 6, including in particular:
    - Activation and termination of vam transmission operation.
    - Determination of the vam generation frequency.
    - Trigger the generation of vam.

    Attributes
    ----------
    vam_coder : vamCoder
        vam Coder object.
    T_Genvam_DCC : int
        Time to wait between vams according to the DCC.
    T_Genvam : int
        Time to wait between vams.
    N_Genvam : int
        Consecutive vams to be generated.
    last_vam_sent : dict
        Last vam sent.
    current_vam_to_send : dict
        Current vam to send.
    """

    def __init__(
        self,
        btp_router: BTPRouter,
        vam_coder: VAMCoder,
        device_data_provider: DeviceDataProvider,
        vru_basic_service_ldm: VRUBasicServiceLDM = None,
    ) -> None:
        """
        Initialize the vam Transmission Management.
        """
        self.logging = logging.getLogger("vru_basic_service")
        self.btp_router: BTPRouter = btp_router
        self.device_data_provider = device_data_provider
        self.vru_basic_service_ldm = vru_basic_service_ldm
        self.vam_coder = vam_coder
        # self.T_Genvam_DCC = T_GenvamMin We don't have a DCC yet.
        self.t_genvam = vam_constants.T_GENVAMMIN
        self.n_genvam = 1
        self.last_vam_sent = None
        self.current_vam_to_send = VAMMessage()
        self.current_vam_to_send.fullfill_with_device_data(self.device_data_provider)

    def location_service_callback(self, tpv: dict) -> None:
        """
        Callback function for location service.

        The VAM gets triggered everytime the location service gets a new position.

        TODO: Conditions 5-7 are not currently implemented. Conditions 5 and 7 will be implemented
            once the Local Dynamic Map (LDM)
        TODO: is available. Condition 6 will be implemented once cluster functionality is
            implemented.

        ETSI TS 103 300-3 V2.2.1 (2023-02) Clause 6.4.1 Individual VAM transmission management
        by VBS at VRU ITS-S.
        The first time individual VAM shall be generated immediately after VBS activation.
        The VAM shall also be generated at the earliest time instant for transmission if any
        of the following conditions are satisfied and the individual VAM transmission does not
        subject to redundancy mitigation techniques:

        1) A VRU is in VRU-IDLE VBS state and has entered VRU-ACTIVE-STANDALONE.
        2) A VRU is in VRU-PASSIVE VBS state; it decides to leave the cluster and enters in
            VRU-ACTIVE-STANDALONE VBS state.
        3) A VRU is in VRU-PASSIVE VBS state; it has determined that VRU cluster leader is lost
            and has decided to enter VRU-ACTIVE-STANDALONE VBS state.
        4) A VRU is in VRU-ACTIVE-CLUSTER-LEADER VBS state; it has determined breaking up the
            cluster and has transmitted VRU cluster VAM with disband indication; it has decided
            to enter VRU-ACTIVE-STANDALONE VBS state.

        Consecutive VAM transmissions are contingent upon the conditions described here.
        Consecutive individual VAM generation events shall occur at an interval equal to or
        larger than T_GenVam. An individual VAM shall be generated for transmission as part of
        a generation event if the originating VRU ITS-S is still in VBS
        VRU-ACTIVE-STANDALONE VBS state, any of the following conditions is satisfied and
        individual VAM transmission is not subject to redundancy mitigation techniques:

        1) The time elapsed since the last time the individual VAM was transmitted exceeds
            T_GENVAMAX.
        2) The Euclidian absolute distance between the current estimated position of the
            reference point of the VRU and the estimated position of the reference point
            lastly included in an individual VAM exceeds a pre-defined threshold
            minReferencePointPositionChangeThreshold.
        3) The difference between the current estimated ground speed of the reference point of
            the VRU and the estimated absolute speed of the reference point of the VRU lastly
            included in an individual VAM exceeds a pre-defined threshold
            minGroundSpeedChangeThreshold.
        4) The difference between the orientation of the vector of the current estimated ground
            velocity of the reference point of the VRU and the estimated orientation of the
            vector of the ground velocity of the reference point of the VRU lastly included in
            an individual VAM exceeds a pre-defined threshold
            minGroundVelocityOrientationChangeThreshold.
        5) The VRU has determined that there is a difference between the current estimated
            trajectory interception probability with vehicle(s) or other VRU(s) and the
            trajectory interception probability with vehicle(s) or other VRU(s) lastly reported
            in an individual VAM exceeds a predefined threshold
            minTrajectoryInterceptionProbChangeThreshold.
        6) The originating ITS-S is a VRU in VRU-ACTIVE-STANDALONE VBS state and has decided
            to join a cluster after its previous individual VAM transmission.
        7) VRU has determined that one or more new vehicles or other VRUs have satisfied the
            following conditions simultaneously after the lastly transmitted VAM:
                - coming closer than Minimum Safe Lateral Distance (MSLaD) laterally;
                - coming closer than Minimum Safe Longitudinal Distance (MSLoD) longitudinally;
                - coming closer than Minimum Safe Vertical Distance (MSVD) vertically.
        """
        self.current_vam_to_send.fullfill_with_tpv_data(tpv)
        self.logging.debug("Fullfilled VAM with TPV data %s", tpv)

        if self.last_vam_sent is None:
            self.send_next_vam()
            return
        received_generation_delta_time = GenerationDeltaTime()
        last_sent_generation_delta_time = GenerationDeltaTime()
        received_generation_delta_time.set_in_normal_timestamp(
            parser.parse(tpv["time"]).timestamp()
        )
        last_sent_generation_delta_time.msec = self.last_vam_sent.vam["vam"][
            "generationDeltaTime"
        ]

        received_position = [int(tpv["lat"]), int(tpv["lon"])]
        last_sent_position = [
            self.last_vam_sent.vam["vam"]["vamParameters"]["basicContainer"][
                "referencePosition"
            ]["latitude"],
            self.last_vam_sent.vam["vam"]["vamParameters"]["basicContainer"][
                "referencePosition"
            ]["longitude"],
        ]

        if (
            received_generation_delta_time - last_sent_generation_delta_time
            >= self.t_genvam
        ):
            self.send_next_vam()
            return
        if (
            Utils.euclidian_distance(received_position, last_sent_position)
            > vam_constants.MINREFERENCEPOINTPOSITIONCHANGETHRESHOLD
        ):
            self.send_next_vam()
            return
        if (
            abs(
                tpv["speed"]
                - self.last_vam_sent.vam["vam"]["vamParameters"][
                    "vruHighFrequencyContainer"
                ]["speed"]["speedValue"]
            )
            > vam_constants.MINGROUNDSPEEDCHANGETHRESHOLD
        ):
            self.send_next_vam()
            return

    def send_next_vam(self) -> None:
        """
        Send the next vam.

        BTP Port Number: 2018
        """
        if self.vru_basic_service_ldm is not None:
            self.vru_basic_service_ldm.add_provider_data_to_ldm(
                self.current_vam_to_send.vam
            )
        request = BTPDataRequest()
        request.btp_type = CommonNH.BTP_B
        request.destination_port = 2018
        request.gn_packet_transport_type = PacketTransportType()
        request.communication_profile = CommunicationProfile.UNSPECIFIED
        request.traffic_class = TrafficClass()
        request.data = self.vam_coder.encode(self.current_vam_to_send.vam)
        request.length = len(request.data)
        self.btp_router.btp_data_request(request)
        self.logging.info(
            "Sent VAM message with timestamp: %s, station_id: %s",
            self.current_vam_to_send.vam['vam']['generationDeltaTime'],
            self.current_vam_to_send.vam['header']['stationId']
        )
        self.last_vam_sent = self.current_vam_to_send
        self.current_vam_to_send = VAMMessage()
        self.current_vam_to_send.fullfill_with_device_data(self.device_data_provider)



from __future__ import annotations
from flexstack.security.ecdsa_backend import ECDSABackend

from flexstack.security.security_coder import SecurityCoder
from .certificate import Certificate, OwnCertificate


class CertificateLibrary:
    """
    Class for managing certificates.

    Attributes
    ----------
    own_certificates : Dict[bytes, OwnCertificate]
        The list of own certificates. (Messages can be signed with those
        certificates) [Key: Hashedid8, Value: OwnCertificate]
    known_authorization_tickets : Dict[bytes, Certificate]
        The list of known authorization tickets. (Messages can be verified
        with those certificates) [Key: Hashedid8, Value: Certificate]
    known_authorization_authorities : Dict[bytes, Certificate]
        The list of known authorization authorities. (Authorization Tickets
        Certificates can be verified with those certificates)
        [Key: Hashedid8, Value: Certificate]
    known_root_certificates : Dict[bytes, Certificate]
        The list of known root certificates. (Authorization Authorities
        Certificates can be verified with those certificates)
        [Key: Hashedid8, Value: Certificate]
    """

    def __init__(
        self,
        root_certificates: list[Certificate],
        aa_certificates: list[Certificate],
        at_certificates: list[Certificate],
    ) -> None:
        """
        Initialize the Certificate Library.

        Parameters
        ----------
        root_certificates : list[Certificate]
            The list of root certificates.
        aa_certificates : list[Certificate]
            The list of authorization authorities certificates.
        at_certificates : list[Certificate]
            The list of authorization tickets certificates.
        """
        self.own_certificates = {}
        self.known_authorization_tickets = {}
        self.known_authorization_authorities = {}
        self.known_root_certificates = {}
        # Root certificates
        for root_certificate in root_certificates:
            self.add_root_certificate(root_certificate)
        # Authorization Authorities certificates
        for aa_certificate in aa_certificates:
            self.add_authorization_authority(aa_certificate)
        # Authorization Tickets certificates
        for at_certificate in at_certificates:
            self.add_authorization_ticket(at_certificate)

    def get_issuer_certificate(self, certificate: Certificate) -> Certificate:
        """
        Get the issuer certificate of a certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to get the issuer certificate from.

        Returns
        -------
        Certificate
            The issuer certificate. It returns None if the issuer is self.
            Or issuer not found.
        """
        if certificate.certificate["issuer"][0] == "self":
            return None
        if certificate.certificate["issuer"][0] == "sha256AndDigest":
            if (
                certificate.certificate["issuer"][1]
                in self.known_root_certificates.keys()
            ):
                return self.known_root_certificates[
                    certificate.certificate["issuer"][1]
                ]
            if (
                certificate.certificate["issuer"][1]
                in self.known_authorization_authorities.keys()
            ):
                return self.known_authorization_authorities[
                    certificate.certificate["issuer"][1]
                ]
            return None
        raise ValueError("Unknown issuer type")

    def add_authorization_ticket(self, certificate: Certificate) -> None:
        """
        Add an authorization ticket to the library.

        Parameters
        ----------
        certificate : Certificate
            The authorization ticket to add.
        """
        if certificate.as_hashedid8() not in self.known_authorization_tickets.keys():
            issuer_certificate = self.get_issuer_certificate(certificate)
            if issuer_certificate is not None:
                if certificate.verify(issuer_certificate):
                    self.known_authorization_tickets[certificate.as_hashedid8()] = (
                        certificate
                    )

    def add_authorization_authority(self, certificate: Certificate) -> None:
        """
        Add an authorization authority to the library.

        Parameters
        ----------
        certificate : Certificate
            The authorization authority to add.
        """
        if (
            certificate.as_hashedid8()
            not in self.known_authorization_authorities.keys()
        ):
            issuer_certificate = self.get_issuer_certificate(certificate)
            if issuer_certificate is not None:
                if certificate.verify(issuer_certificate):
                    self.known_authorization_authorities[certificate.as_hashedid8()] = (
                        certificate
                    )

    def add_root_certificate(self, certificate: Certificate) -> None:
        """
        Add a root certificate to the library.

        Parameters
        ----------
        certificate : Certificate
            The root certificate to add.
        """
        if certificate.verify():
            self.known_root_certificates[certificate.as_hashedid8()] = certificate

    def add_own_certificate(self, certificate: OwnCertificate) -> None:
        """
        Add a certificate to the library.

        Parameters
        ----------
        certificate : OwnCertificate
            The certificate to add
        """
        issuer_certificate = self.get_issuer_certificate(certificate)
        if issuer_certificate is not None and certificate.verify(issuer_certificate):
            self.own_certificates[certificate.as_hashedid8()] = certificate

    def get_authorization_ticket_by_hashedid8(self, hashedid8: bytes) -> Certificate:
        """
        Get an authorization ticket by its hashedid8.

        Parameters
        ----------
        hashedid8 : bytes
            The hashedid8 of the authorization ticket.
        """
        if hashedid8 in self.known_authorization_tickets.keys():
            return self.known_authorization_tickets[hashedid8]
        return None

    def verify_sequence_of_certificates(
        self, certificates: list[dict], coder: SecurityCoder, backend: ECDSABackend
    ) -> Certificate:
        """
        Verification of a sequence of certificates as specified in IEEE 1609.2. Signer Identifer.
        The first certificate is the authorization ticket. And then each one is the issuer of the one before.

        Parameters
        ----------
        certificates : list[dict]
            The sequence of certificates to verify.
        coder : SecurityCoder
            The coder to use for encoding and decoding.
        backend : ECDSABackend
            The backend to use for ECDSA operations.

        Returns
        -------
        Certificate
            The Authorization ticket if verified. Returns none otherwise.
        """
        if len(certificates) == 0:
            return None
        elif len(certificates) == 1:
            temp_certificate = Certificate(coder=coder, backend=backend)
            temp_certificate.from_dict(certificates[0])
            if (
                temp_certificate.as_hashedid8()
                in self.known_authorization_tickets.keys()
            ):
                return self.known_authorization_tickets[temp_certificate.as_hashedid8()]
            issuer_certificate = self.get_issuer_certificate(temp_certificate)
            if issuer_certificate is not None and temp_certificate.verify(
                issuer_certificate
            ):
                self.add_authorization_ticket(temp_certificate)
                return temp_certificate
        elif len(certificates) == 2:
            authorization_ticket = Certificate(coder=coder, backend=backend)
            authorization_ticket.from_dict(certificates[0])
            authorization_authority = Certificate(coder=coder, backend=backend)
            authorization_authority.from_dict(certificates[1])
            authorization_authority_issuer_hashedid8 = (
                authorization_authority.get_issuer_hashedid8()
            )
            if (
                authorization_authority_issuer_hashedid8
                in self.known_root_certificates.keys()
            ):
                root_certificate = self.known_root_certificates[
                    authorization_authority_issuer_hashedid8
                ]
                if authorization_authority.verify(root_certificate):
                    self.add_authorization_authority(authorization_authority)
                    if authorization_ticket.verify(authorization_authority):
                        self.add_authorization_ticket(authorization_ticket)
                        return authorization_ticket
        elif len(certificates) == 3:
            root_certificate = Certificate(coder=coder, backend=backend)
            root_certificate.from_dict(certificates[-1])
            if root_certificate.as_hashedid8() in self.known_root_certificates.keys():
                return self.verify_sequence_of_certificates(
                    certificates[:-1], coder, backend
                )
        return None



from enum import Enum
from base64 import b64encode, b64decode
from .position_vector import LongPositionVector
from ..security.security_profiles import SecurityProfile


class CommonNH(Enum):
    """
    Common Next Header class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.7

    Attributes
    ----------
    ANY :
        Any Next Header.
    BTP-A :
        BTP-A Next Header.
    BTP-B :
        BTP-B Next Header.
    IPV6 :
        IPv6 Next Header.
    """

    ANY = 0
    BTP_A = 1
    BTP_B = 2
    IPV6 = 3


class HeaderType(Enum):
    """
    Header Type class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.7.4

    Attributes
    ----------
    ANY :
        Any Header Type.
    BEACON :
        Beacon Header Type.
    GEOUNICAST :
        GeoUnicast Header Type.
    GEOANYCAST :
        Geographically-Scoped Anycast (GAC) Header Type.
    GEOBROADCAST :
        Geographically-Scoped broadcast (GBC)  Header Type.
    TSB :
        Topologically-scoped broadcast (TSB) Header Type.
    LS :
        Location Service Header Type.
    """

    ANY = 0
    BEACON = 1
    GEOUNICAST = 2
    GEOANYCAST = 3
    GEOBROADCAST = 4
    TSB = 5
    LS = 6


class HeaderSubType(Enum):
    """
    Common Header Subtype class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01).
    Section 9.7.5
    """

    UNSPECIFIED = 0


class GeoAnycastHST(Enum):
    """
    Geographically-Scoped Anycast (GAC) Header Subtype class. As specified in
    ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.7.4

    Attributes
    ----------
    GEOANYCAST_CIRCLE :
        Geographically-Scoped Anycast (GAC) Circle Header Subtype.
    GEOANYCAST_RECT :
        Geographically-Scoped Anycast (GAC) Rectangle Header Subtype.
    GEOANYCAST_ELIP :
        Geographically-Scoped Anycast (GAC) Ellipse Header Subtype.
    """

    GEOANYCAST_CIRCLE = 0
    GEOANYCAST_RECT = 1
    GEOANYCAST_ELIP = 2


class GeoBroadcastHST(Enum):
    """
    Geographically-Scoped broadcast (GBC) Header Subtype class.
    As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01).
    Section 9.7.4

    Attributes
    ----------
    GEOBROADCAST_CIRCLE :
        Geographically-Scoped broadcast (GBC) Circle Header Subtype.
    GEOBROADCAST_RECT :
        Geographically-Scoped broadcast (GBC) Rectangle Header Subtype.
    GEOBROADCAST_ELIP :
        Geographically-Scoped broadcast (GBC) Ellipse Header Subtype.
    """

    GEOBROADCAST_CIRCLE = 0
    GEOBROADCAST_RECT = 1
    GEOBROADCAST_ELIP = 2


class TopoBroadcastHST(Enum):
    """
    Topologically-scoped broadcast (TSB) Header Subtype class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01).
    Section 9.7.4

    Attributes
    ----------
    SINGLE_HOP :
        Single Hop Header Subtype.
    MULTI_HOP :
        Multi Hop Header Subtype.
    """

    SINGLE_HOP = 0
    MULTI_HOP = 1


class LocationServiceHST(Enum):
    """
    Location Service Header Subtype class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.7.4

    Attributes
    ----------
    LS_REQUEST :
        Location Service Request Header Subtype.
    LS_REPLY :
        Location Service Reply Header Subtype.
    """

    LS_REQUEST = 0
    LS_REPLY = 1


class TrafficClass:
    """
    Common Traffic class class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.7.5

    Attributes
    ----------
    scf : bool
        (1 bit) Store Carry Forward (SCF) flag. Indicates whether the packet shall be buffered when no suitable
        neighbour exists
    channel_offload : bool
        (1 bit) Channel Offload flag. Indicates whether the packet may be offloaded to another channel than
        specified in the TC ID
    tc_id : int
        (6 bit unsigned integer) Traffic class identifier. TC ID as specified in the media-dependent part of
        GeoNetworking corresponding to the interface over which the packet will be transmitted
    """

    def __init__(self) -> None:
        self.scf = False
        self.channel_offload = False
        self.tc_id = 0

    def set_scf(self, scf: bool) -> None:
        """
        Set the SCF flag.

        Parameters
        ----------
        scf : bool
            SCF flag.
        """
        self.scf = scf

    def set_tc_id(self, tc_id: int) -> None:
        """
        Set the traffic class identifier.

        Parameters
        ----------
        tc_id : int
            Traffic class identifier.
        """
        if tc_id < 0 or tc_id > 63:
            raise ValueError("Traffic class identifier must be between 0 and 63")
        self.tc_id = tc_id

    def set_channel_offload(self, channel_offload: bool) -> None:
        """
        Set the channel offload flag.

        Parameters
        ----------
        channel_offload : bool
            Channel offload flag.
        """
        self.channel_offload = channel_offload

    def encode_to_int(self) -> int:
        """
        Encodes the traffic class to an integer.

        Returns
        -------
        int :
            Encoded traffic class. 1 byte.
        """
        return (self.scf << 7) | (self.channel_offload << 6) | self.tc_id

    def encode_to_bytes(self) -> bytes:
        """
        Encodes the traffic class to bytes.

        Returns
        -------
        bytes :
            Encoded traffic class. 1 byte.
        """
        return self.encode_to_int().to_bytes(1, "big")

    def decode_from_int(self, tc: int) -> None:
        """
        Decodes the traffic class from an integer.

        Parameters
        ----------
        tc : int
            Encoded traffic class. 1 byte.
        """
        self.scf = (tc >> 7) & 1
        self.channel_offload = (tc >> 6) & 1
        self.tc_id = tc & 63

    def decode_from_bytes(self, tc: bytes) -> None:
        """
        Decodes the traffic class from a byte array.

        Parameters
        ----------
        tc : bytes
            Byte array containing the traffic class.
        """
        self.decode_from_int(int.from_bytes(tc, "big"))

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, TrafficClass):
            return (
                self.scf == __o.scf
                and self.channel_offload == __o.channel_offload
                and self.tc_id == __o.tc_id
            )
        return False


class PacketTransportType:
    """
    Packet Transport Type class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex J2
    Uses the Header Type and Header Subtype fields from the Common Header. As specified in ETSI EN 302 636-4-1 V1.4.1
    (2020-01). Section 9.7.4 Table 9

    Attributes
    ----------
    header_type : HeaderType
        Header Type.
    header_subtype : Enum
        Header Subtype.
    """

    def __init__(self) -> None:
        """
        Initialize the Packet Transport Type.

        By default now is set to Single Hop Broadcast.
        """
        self.header_type = HeaderType.TSB
        self.header_subtype = TopoBroadcastHST.SINGLE_HOP

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the Packet Transport Type.

        Returns
        -------
        dict :
            Dictionary representation of the Packet Transport Type.
        """
        return {
            "header_type": self.header_type.value,
            "header_subtype": self.header_subtype.value,
        }

    def from_dict(self, packet_transport_type: dict) -> None:
        """
        Initialize the Packet Transport Type from a dictionary.

        Parameters
        ----------
        packet_transport_type : dict
            Dictionary containing the Packet Transport Type.
        """
        self.header_type = HeaderType(packet_transport_type["header_type"])
        if self.header_type == HeaderType.GEOANYCAST:
            self.header_subtype = GeoAnycastHST(packet_transport_type["header_subtype"])
        elif self.header_type == HeaderType.GEOBROADCAST:
            self.header_subtype = GeoBroadcastHST(
                packet_transport_type["header_subtype"]
            )
        elif self.header_type == HeaderType.TSB:
            self.header_subtype = TopoBroadcastHST(
                packet_transport_type["header_subtype"]
            )
        elif self.header_type == HeaderType.LS:
            self.header_subtype = LocationServiceHST(
                packet_transport_type["header_subtype"]
            )
        else:
            self.header_subtype = HeaderSubType(packet_transport_type["header_subtype"])


class CommunicationProfile(Enum):
    """
    Communication Profile class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex J2

    Attributes
    ----------
    UNSPECIFIED (0) :
        Unspecified
    """

    UNSPECIFIED = 0


class Area:
    """
    Area class. Not specified in the standard


    Attributes
    ----------
    latitude : int
        Latitude of the center of the area. In 1/10 micro degree.
    longitude : int
        Longitude of the center of the area. In 1/10 micro degree.
    a : int
        Length of the semi-major axis. In meters.
    b : int
        Length of the semi-minor axis. In meters.
    angle : int
        In degrees from North
    """

    def __init__(self) -> None:
        """
        Initialize the Area.
        """
        self.latitude = 0
        self.longitude = 0
        self.a = 0
        self.b = 0
        self.angle = 0

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the Area.

        Returns
        -------
        dict :
            Dictionary representation of the Area.
        """
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "a": self.a,
            "b": self.b,
            "angle": self.angle,
        }

    def from_dict(self, area: dict) -> None:
        """
        Initialize the Area from a dictionary.

        Parameters
        ----------
        area : dict
            Dictionary containing the area.
        """
        self.latitude = area["latitude"]
        self.longitude = area["longitude"]
        self.a = area["a"]
        self.b = area["b"]
        self.angle = area["angle"]


class GNDataRequest:
    """
    GN Data Request class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex J2.2


    Attributes
    ----------
    upper_protocol_entity : CommonNH
        Upper Protocol Entity.
    packet_transport_type : PacketTransportType
        Packet Transport Type.
    communication_profile : CommunicationProfile
        Communication Profile.
    security_profile : SecurityProfile
        Security Profile.
    its_aid : int
        ITS AID.
    security_permissions : bytes
        Security Permissions.
    traffic_class : TrafficClass
        Traffic Class.
    length : int
        Length of the payload.
    data : bytes
        Payload.
    area : Area
        Area of the GBC algorithm. Only used when the packet transport type is GBC.

    THIS CLASS WILL BE EXTENDED WHEN FURTHER PACKET TYPES ARE IMPLEMENTED
    """

    def __init__(self) -> None:
        """
        Initialize the GN Data Request.
        """
        self.upper_protocol_entity = CommonNH.ANY
        self.packet_transport_type = PacketTransportType()
        self.communication_profile = CommunicationProfile.UNSPECIFIED
        self.security_profile = SecurityProfile.NO_SECURITY
        self.its_aid = 0
        self.security_permissions = b"\x00"
        self.traffic_class = TrafficClass()
        self.length = 0
        self.data = b""
        self.area = Area()

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the GN Data Request.

        Returns
        -------
        dict :
            Dictionary representation of the GN Data Request.
        """
        return {
            "upper_protocol_entity": self.upper_protocol_entity.value,
            "packet_transport_type": self.packet_transport_type.to_dict(),
            "communication_profile": self.communication_profile.value,
            "traffic_class": b64encode(self.traffic_class.encode_to_bytes()).decode(
                "utf-8"
            ),
            "length": self.length,
            "data": b64encode(self.data).decode("utf-8"),
            "area": self.area.to_dict(),
        }

    def from_dict(self, gn_data_request: dict) -> None:
        """
        Initialize the GN Data Request from a dictionary.

        Parameters
        ----------
        gn_data_request : dict
            Dictionary containing the GN Data Request.
        """
        self.upper_protocol_entity = CommonNH(gn_data_request["upper_protocol_entity"])
        self.packet_transport_type.from_dict(gn_data_request["packet_transport_type"])
        self.communication_profile = CommunicationProfile(
            gn_data_request["communication_profile"]
        )
        self.traffic_class.decode_from_bytes(
            b64decode(gn_data_request["traffic_class"])
        )
        self.length = gn_data_request["length"]
        self.data = b64decode(gn_data_request["data"])
        self.area.from_dict(gn_data_request["area"])


class ResultCode(Enum):
    """
    Result Code class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex J3

    Attributes
    ----------
    ACCEPTED (1) :
        Accepted
    MAXIMUM_LENGTH_EXCEEDED (2) :
        The size of the T/GN6-PDU exceeds the GN protocol constant itsGnMaxSduSize;
    MAXIMUM_LIFETIME_EXCEEDED (3) :
        The lifetime exceeds the maximum value of the GN protocol constant itsGnMaxPacketLifetime;
    REPETITION_INTERVAL_TOO_SMALL (4) :
        The repetition interval is too small;
    UNSUPPORTED_TRAFFIC_CLASS (5) :
        The traffic class is not supported;
    GEOGRAPHICAL_SCOPE_TOO_LARGE (6) :
        The geographical scope is too large;
    UNSPECIFIED (7) :
        Unspecified
    """

    ACCEPTED = 1
    MAXIMUM_LENGTH_EXCEEDED = 2
    MAXIMUM_LIFETIME_EXCEEDED = 3
    REPETITION_INTERVAL_TOO_SMALL = 4
    UNSUPPORTED_TRAFFIC_CLASS = 5
    GEOGRAPHICAL_SCOPE_TOO_LARGE = 6
    UNSPECIFIED = 7


class GNDataConfirm:
    """
    GN Data Confirm class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex J3

    Attributes
    ----------
    result_code : ResultCode
        Result Code.

    """

    def __init__(self) -> None:
        """
        Initialize the GN Data Confirm.
        """
        self.result_code = ResultCode.UNSPECIFIED


class GNDataIndication:
    """
    GN Data Indication class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex J4

    Attributes
    ----------
    upper_protocol_entity : CommonNH
        Upper Protocol Entity.
    packet_transport_type : PacketTransportType
        Packet Transport Type.
    source_position_vector : LongPositionVector
        Source Position Vector.
    traffic_class : TrafficClass
        Traffic Class.
    length : int
        Length of the payload.
    data : bytes
        Payload.
    """

    def __init__(self) -> None:
        """
        Initialize the GN Data Indication.
        """
        self.upper_protocol_entity = CommonNH.ANY
        self.packet_transport_type = PacketTransportType()
        self.source_position_vector = LongPositionVector()
        self.traffic_class = TrafficClass()
        self.length = 0
        self.data = b""

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the GN Data Indication.

        Returns
        -------
        dict :
            Dictionary representation of the GN Data Indication.
        """
        return {
            "upper_protocol_entity": self.upper_protocol_entity.value,
            "packet_transport_type": self.packet_transport_type.to_dict(),
            "source_position_vector": b64encode(
                self.source_position_vector.encode()
            ).decode("utf-8"),
            "traffic_class": b64encode(self.traffic_class.encode_to_bytes()).decode(
                "utf-8"
            ),
            "length": self.length,
            "data": b64encode(self.data).decode("utf-8"),
        }

    def from_dict(self, gn_data_indication: dict) -> None:
        """
        Initialize the GN Data Indication from a dictionary.

        Parameters
        ----------
        gn_data_indication : dict
            Dictionary containing the GN Data Indication.
        """
        self.upper_protocol_entity = CommonNH(
            gn_data_indication["upper_protocol_entity"]
        )
        self.packet_transport_type.from_dict(
            gn_data_indication["packet_transport_type"]
        )
        self.source_position_vector.decode(
            b64decode(gn_data_indication["source_position_vector"])
        )
        self.traffic_class.decode_from_bytes(
            b64decode(gn_data_indication["traffic_class"])
        )
        self.length = gn_data_indication["length"]
        self.data = b64decode(gn_data_indication["data"])



class LDMMaintenanceKeyError(Exception):
    """
    Exception raised when LDM Maintenance Key Error occurs
    """


class LDMServiceKeyError(Exception):
    """
    Exception raised when LDM Service Key Error occurs
    """


class LDMDatabaseKeyError(Exception):
    """
    Excepetion raised when LDM Database Key Error occurs
    """



from enum import Enum


class SecurityProfile(Enum):
    """
    Security profile of a message.
    """
    NO_SECURITY = 0
    COOPERATIVE_AWARENESS_MESSAGE = 1
    DECENTRALIZED_ENVIRONMENTAL_NOTIFICATION_MESSAGE = 2



from enum import Enum
from .gn_address import GNAddress


class LocalGnAddrConfMethod(Enum):
    """
    LocalGnAddrConfMethod. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    AUTO (0) :
        Local GN_ADDR is configured from MIB
    MANAGED (1) :
        Local GN_ADDR is configured via the GN management using the service primitive GN-MGMT (annex K)
    ANONYMOUS (2) : Local GN_ADDR is configured by the Security entity
    """

    AUTO = 0
    MANAGED = 1
    ANONYMOUS = 2


class GnIsMobile(Enum):
    """
    GnIsMobile. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    STATIONARY (0) :
        The GeoAdhoc router is stationary
    MOBILE (1) :
        The GeoAdhoc router is mobile
    """

    STATIONARY = 0
    MOBILE = 1


class GnIfType(Enum):
    """
    GnIfType. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    UNSPECIFIED (0) :
        The interface type is unspecified
    ITS-G5 (1) :
        The interface is an ITS-G5 interface
    LTE-V2X (2) :
        The interface is an LTE-V2X interface
    """

    UNSPECIFIED = 0
    ITS_G5 = 1
    LTE_V2X = 2


class GnSecurity(Enum):
    """
    GnSecurity. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    DISABLED (0) :
        Security is disabled
    ENABLED (1) :
        Security is enabled
    """

    DISABLED = 0
    ENABLED = 1


class SnDecapResultHandling(Enum):
    """
    SnDecapResultHandling. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    STRICT (0) :
        The packet is dropped
    NON-STRICT (1) :
        The packet is forwarded
    """

    STRICT = 0
    NON_STRICT = 1


class NonAreaForwardingAlgorithm(Enum):
    """
    NonAreaForwardingAlgorithm. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    UNSPECIFIED (0) :
        The forwarding algorithm is unspecified
    GREEDY (1) :
        Default forwarding algorithm outside target area
    """

    UNSPECIFIED = 0
    GREEDY = 1


class AreaForwardingAlgorithm(Enum):
    """
    AreaForwardingAlgorithm. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex H.

    Attributes
    ----------
    UNSPECIFIED (0) :
        The forwarding algorithm is unspecified
    SIMPLE (1) :
        The simple forwarding algorithm inside target area
    CBF (1) :
        Default forwarding algorithm inside target area
    """

    UNSPECIFIED = 0
    SIMPLE = 1
    CBF = 2


class MIB:
    # pylint: disable=too-many-instance-attributes, invalid-name
    """Management Information Base (MIB) for GeoNetworking. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01).
    Annex H.

    Attributes
    ----------
    itsGnLocalGnAddr : GNAddress
        GeoNetworking address of the GeoAdhoc router
    itsGnLocalGnAddrConfMethod : LocalGnAddrConfMethod
        Configuration method for the GeoNetworking address of the GeoAdhoc router
    itsGnProtocolVersion : int
        GeoNetworking protocol version
    itsGnIsMobile : GnIsMobile
        Mobility status of the GeoAdhoc router
    itsGnIfType : GnIfType
        Interface type
    itsGnMinUpdateFrequencyEPV : int
        Minimum update frequency of EPV in [1/ms]
    itsGnPaiInterval : int
        Distance related to the confidence interval for latitude and longitude [m].
    itsGnMaxSduSize : int
        Maximum size of an SDU in [byte].
    itsGnMaxGeoNetworkingHeaderSize : int
        Maximum size of a GeoNetworking header in [byte].
    itsGnLifetimeLocTE : int
        Lifetime of location table entry [s]
    itsGnSecurity : GnSecurity
        Security status
    itsGnSnDecapResultHandling : SnDecapResultHandling
        Handling of the result of the security decapsulation
    itsGnLocationServiceMaxRetrans : int
        Maximum number of retransmissions for location service requests
    itsGnLocationServiceRetransmitTimer : int
        Duration of Location service retransmit timer [ms]
    itsGnLocationServicePacketBufferSize : int
        Size of Location service packet buffer [Octets]
    itsGnBeaconServiceRetransmitTimer : int
        Duration of Beacon service retransmit timer [ms]
    itsGnBeaconServiceMaxJitter : int
        Maximum jitter for Beacon service retransmission [ms]
    itsGnDefaultHopLimit : int
        Default hop limit
    itsGnDPLLength : int
        Length of Duplicate Packet List (DPL) per source
    itsGnMaxPacketLifetime : int
        Maximum packet lifetime [s]
    itsGnDefaultPacketLifetime : int
        Default packet lifetime [s]
    itsGnMaxPacketDataRate : int
        Maximum packet data rate [ko/s]
    itsGnMaxPacketDataRateEmaBeta : int
        Weight factor for the Exponential Moving Average of the packet data rate PDR in percent
    itsGnMaxGeoAreaSize : int
        Maximum size of the geographical area for a GBC and GAC packet [km2]
    itsGnMinPacketRepetitionInterval : int
        Lower limit of the packet repetition interval [ms]
    itsGnNonAreaForwardingAlgorithm : NonAreaForwardingAlgorithm
        Forwarding algorithm outside target area
    itsGnAreaForwardingAlgorithm : AreaForwardingAlgorithm
        Forwarding algorithm inside target area
    itsGnCbfMinTime : int
        Minimum duration a GN packet shall be buffered in the CBF packet buffer [ms]
    itsGnCbfMaxTime : int
        Maximum duration a GN packet shall be buffered in the CBF packet buffer [ms]
    itsGnDefaultMaxCommunicationRange : int
        Default maximum communication range [m]
    itsGnBroadcastCBFDefSectorAngle : int
        Default threshold angle for advanced GeoBroadcast algorithm in clause F.4 [degrees]
    itsGnUcForwardingPacketBufferSize : int
        Size of UC forwarding packet buffer [Ko]
    itsGnBcForwardingPacketBufferSize : int
        Size of BC forwarding packet buffer [Ko]
    itsGnCbfPacketBufferSize : int
        Size of CBF packet buffer [Ko]
    itsGnDefaultTrafficClass : int
        Default traffic class
    """

    def __init__(self) -> None:
        self.itsGnLocalGnAddr = GNAddress()
        # self.itsGnLocalGnAddrConfMethod = LocalGnAddrConfMethod.MANAGED
        self.itsGnLocalGnAddrConfMethod = LocalGnAddrConfMethod.AUTO
        self.itsGnProtocolVersion = 1
        self.itsGnIsMobile = GnIsMobile.MOBILE
        self.itsGnIfType = GnIfType.UNSPECIFIED
        self.itsGnMinUpdateFrequencyEPV = 1000
        self.itsGnPaiInterval = 80
        self.itsGnMaxSduSize = 1398
        self.itsGnMaxGeoNetworkingHeaderSize = 88
        self.itsGnLifetimeLocTE = 20
        self.itsGnSecurity = GnSecurity.DISABLED
        self.itsGnSnDecapResultHandling = SnDecapResultHandling.STRICT
        self.itsGnLocationServiceMaxRetrans = 10
        self.itsGnLocationServiceRetransmitTimer = 1000
        self.itsGnLocationServicePacketBufferSize = 1024
        self.itsGnBeaconServiceRetransmitTimer = 3000
        self.itsGnBeaconServiceMaxJitter = self.itsGnBeaconServiceRetransmitTimer / 4
        self.itsGnDefaultHopLimit = 10
        self.itsGnDPLLength = 8
        self.itsGnMaxPacketLifetime = 600
        self.itsGnDefaultPacketLifetime = 60
        self.itsGnMaxPacketDataRate = 100
        self.itsGnMaxPacketDataRateEmaBeta = 90
        self.itsGnMaxGeoAreaSize = 10
        self.itsGnMinPacketRepetitionInterval = 100
        self.itsGnNonAreaForwardingAlgorithm = NonAreaForwardingAlgorithm.GREEDY
        self.itsGnAreaForwardingAlgorithm = AreaForwardingAlgorithm.CBF
        self.itsGnCbfMinTime = 1
        self.itsGnCbfMaxTime = 100
        self.itsGnDefaultMaxCommunicationRange = 1000
        self.itsGnBroadcastCBFDefSectorAngle = 30
        self.itsGnUcForwardingPacketBufferSize = 256
        self.itsGnBcForwardingPacketBufferSize = 1024
        self.itsGnCbfPacketBufferSize = 256
        self.itsGnDefaultTrafficClass = 0



from __future__ import annotations
from copy import deepcopy, copy
from hashlib import sha256

from .security_coder import SecurityCoder
from .ecdsa_backend import ECDSABackend


class Certificate:
    """
    Class to handle Certificates as specified in ETSI TS 103 097 V2.1.1 (2021-10) standard.

    Attributes
    ----------

    """

    def __init__(self, coder: SecurityCoder, backend: ECDSABackend):
        """
        Initialize the Certificate.

        Parameters
        ----------
        coder : SecurityCoder
            Security Coder.
        backend : ECDSABackend
            ECDSA Backend.
        """
        self.coder: SecurityCoder = coder
        self.backend: ECDSABackend = backend
        self.verifier: CertificateVerifier = CertificateVerifier(coder, backend)
        self.certificate: dict = None
        self.issuer: "Certificate" = None
        self.verified: bool = False

    def from_dict(self, certificate: dict, issuer: "Certificate" = None) -> None:
        """
        Sets up the certificate from a dictionary.

        Parameters
        ----------
        certificate : dict
            The certificate to be set up.
        issuer : Certificate
            The issuer of the certificate
        """
        self.certificate = deepcopy(certificate)
        self.verified = self.verify(issuer)

    def decode(self, certificate: bytes, issuer: "Certificate" = None) -> None:
        """
        Decode the certificate and set it up in the Certificate object.

        Parameters
        ----------
        certificate : bytes
            The certificate to be decoded.
        issuer : Certificate
            The issuer of the certificate
        """
        self.from_dict(
            self.coder.decode_etsi_ts_103097_certificate(certificate), issuer
        )

    def as_hashedid8(self) -> bytes:
        """
        Return the certificate as a hashedid8.

        Returns
        -------
        bytes
            The certificate as a hashedid8.
        """
        m = sha256()
        m.update(self.coder.encode_etsi_ts_103097_certificate(self.certificate))
        return m.digest()[-8:]

    def encode(self) -> bytes:
        """
        Return the certificate as bytes.

        Returns
        -------
        bytes
            The certificate as bytes.
        """
        return self.coder.encode_etsi_ts_103097_certificate(self.certificate)

    def get_list_of_its_aid(self) -> list[int]:
        """
        Return the list of ITS AID.

        Returns
        -------
        list[int]
            The list of ITS AID.
        """
        to_return = []
        for psid_ssp in self.certificate["toBeSigned"]["appPermissions"]:
            to_return.append(psid_ssp["psid"])
        return to_return

    def as_clear_certificate(self):
        """
        Generate a White Certificate and set it up to the Certificate object.
        """
        self.certificate = {
            "version": 3,
            "type": "explicit",
            "issuer": (
                "sha256AndDigest",
                (0xA495991B7852B855).to_bytes(8, byteorder="big"),
            ),
            "toBeSigned": {
                "id": ("name", "i2cat.net"),
                "cracaId": (0xA49599).to_bytes(3, byteorder="big"),
                "crlSeries": 0,
                "validityPeriod": {"start": 0, "duration": ("seconds", 30)},
                "appPermissions": [
                    {
                        "psid": 0,
                    }
                ],
                "certIssuePermissions": [
                    {
                        "subjectPermissions": ("all", None),
                        "minChainLength": 1,
                        "chainLengthRange": 0,
                        "eeType": (b"\x00", 1),
                    }
                ],
                "verifyKeyIndicator": (
                    "verificationKey",
                    ("ecdsaNistP256", ("fill", None)),
                ),
            },
            "signature": (
                "ecdsaNistP256Signature",
                {
                    "rSig": ("fill", None),
                    "sSig": (0xA495991B7852B855).to_bytes(32, byteorder="big"),
                },
            ),
        }

    def get_issuer_hashedid8(self) -> bytes:
        """
        Returns the issuer HashedId8 stored in the dict.

        Returns
        -------
        bytes
            The issuer HashedId8. Is none if it's self signed

        Raises
        ------
        ValueError
            If the issuer type is unknown
        """
        if self.certificate["issuer"][0] == "sha256AndDigest":
            return self.certificate["issuer"][1]
        if self.certificate["issuer"][0] == "self":
            return None
        raise ValueError("Unknown issuer type")

    def verify(self, issuer: "Certificate" = None) -> bool:
        """
        Verify the certificate.

        Parameters
        ----------
        issuer : Certificate
            The issuer of the certificate.
        """
        return self.verifier.verify(self, issuer)

    def __str__(self):
        """
        Return the certificate as a string.

        Returns
        -------
        str
            The certificate as a string.
        """
        return str(self.certificate["toBeSigned"]["id"][1])


class OwnCertificate(Certificate):
    """
    Class that handles certificates that are generated by the user. And thus it has the private key.

    Attributes
    ----------
    key_id : int
        The key id of the pair of keys used to sign the certificate.
    certificate_issuer : CertificateIssuer
        The certificate issuer.
    """

    def __init__(self, coder: SecurityCoder, backend: ECDSABackend):
        """
        Initialize the OwnCertificate.

        Parameters
        ----------
        coder : SecurityCoder
            Security Coder.
        backend : ECDSABackend
            ECDSA Backend.
        """
        super().__init__(coder, backend)
        self.key_id = self.backend.create_key()
        self.certificate_issuer = CertificateIssuer(coder, backend)
        self.as_clear_certificate()

    def initialize_certificate(
        self, to_be_signed_certificate: dict, issuer: "OwnCertificate" = None
    ) -> None:
        """
        Initializes the certificate.

        Parameters
        ----------
        to_be_signed_certificate : dict
            The to be signed certificate to be initialized.
        issuer : OwnCertificate
            The issuer of the certificate. If None, the certificate will be self signed.
        """
        if self.verify_to_be_signed_certificate(to_be_signed_certificate):
            self.certificate["toBeSigned"] = deepcopy(to_be_signed_certificate)
            self.certificate["toBeSigned"]["verifyKeyIndicator"] = (
                "verificationKey",
                self.backend.get_public_key(self.key_id),
            )
            if issuer and issuer.verify(issuer.issuer):
                issuer.issue_certificate(self)
            elif issuer is None:
                self.issue_certificate(self)

    def verify_to_be_signed_certificate(self, to_be_signed_certificate: dict) -> bool:
        """
        Verifies if a to be signed certificate is valid.

        Parameters
        ----------
        to_be_signed_certificate : dict
            The to be signed certificate to be verified.

        Returns
        -------
        bool
            True if the to be signed certificate is valid, False otherwise.
        """
        try:
            self.coder.encode_ToBeSignedCertificate(to_be_signed_certificate)
            return True
        except Exception:
            return False

    def issue_certificate(self, certificate: Certificate) -> None:
        """
        Issue a certificate to the given certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be issued.
        """
        self.certificate_issuer.issue_certificate(certificate, self)

    def sign_message(self, message: bytes) -> bytes:
        """
        Sign a message with the private key.

        Parameters
        ----------
        message : bytes

        Returns
        -------
        bytes
            The signed message.
        """
        return self.backend.sign(message, self.key_id)


class CertificateVerifier:
    """
    Class that verifies certificates.

    Attributes
    ----------
    coder : SecurityCoder
        Security Coder.
    backend : ECDSABackend
        ECDSA Backend.
    """

    def __init__(self, coder: SecurityCoder, backend: ECDSABackend):
        """
        Initialize the Certificate Verifier.

        Parameters
        ----------
        coder : SecurityCoder
            Security Coder.
        backend : ECDSABackend
            ECDSA Backend.
        """
        self.coder: SecurityCoder = coder
        self.backend: ECDSABackend = backend

    def signature_is_nist_p256(self, certificate: Certificate) -> bool:
        """
        Check if the signature is NISTP256.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the signature is NISTP256, False otherwise.
        """
        return certificate.certificate["signature"][0] == "ecdsaNistP256Signature"

    def verification_key_is_nist_p256(self, certificate: Certificate) -> bool:
        """
        Check if the verification key is NISTP256.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the verification key is NISTP256, False otherwise.
        """
        return (
            certificate.certificate["toBeSigned"]["verifyKeyIndicator"][0]
            == "verificationKey"
            and certificate.certificate["toBeSigned"]["verifyKeyIndicator"][1][0]
            == "ecdsaNistP256"
        )

    def certificate_is_self_signed(self, certificate: Certificate) -> bool:
        """
        Check if the certificate is self signed.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the certificate is self signed, False otherwise.
        """
        return (
            certificate.certificate["issuer"][0] == "self"
            and certificate.certificate["issuer"][1] == "sha256"
        )

    def certificate_is_issued(self, certificate: Certificate) -> bool:
        """
        Check if the certificate is issued.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the certificate is issued, False otherwise.
        """
        return (
            certificate.certificate["issuer"][0] == "sha256AndDigest"
            and len(certificate.certificate["issuer"][1]) == 8
        )

    def check_corresponding_issuer(
        self, certificate: Certificate, issuer: Certificate
    ) -> bool:
        """
        Check if the issuer corresponds to the certificate stated issuer.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the issuer corresponds to the certificate stated issuer, False otherwise
        """
        return certificate.certificate["issuer"][1] == issuer.as_hashedid8()

    def verify_signature(
        self, to_be_signed_certificate: dict, signature: dict, verification_key: dict
    ) -> bool:
        """
        Verify the signature of a certificate.

        Parameters
        ----------
        to_be_signed_certificate : dict
            The to be signed certificate to be verified.
        signature : dict
            The signature to be verified.
        verification_key : dict
            The verification key to be used.

        Returns
        -------
        bool
            True if the signature is valid, False otherwise.
        """
        try:
            return self.backend.verify_with_pk(
                self.coder.encode_ToBeSignedCertificate(to_be_signed_certificate),
                signature,
                verification_key,
            )
        except Exception:
            return False

    def verify_self_signed_certificate(self, certificate: Certificate) -> bool:
        """
        Verifies a self signed Certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be verified.

        Returns
        -------
        bool
            True if the certificate is valid, False otherwise.
        """
        if self.certificate_is_self_signed(certificate):
            if self.signature_is_nist_p256(
                certificate
            ) and self.verification_key_is_nist_p256(certificate):
                if self.verify_signature(
                    certificate.certificate["toBeSigned"],
                    certificate.certificate["signature"],
                    certificate.certificate["toBeSigned"]["verifyKeyIndicator"][1],
                ):
                    certificate.verified = True
                    return True
        return False

    def verify_correct_issuing_permissions(
        self, certificate: Certificate, issuer: Certificate
    ) -> bool:
        """
        Verifies if a certificate has the correct issuing permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be verified.
        issuer : Certificate
            The issuer of the certificate.

        Returns
        -------
        bool
            True if the certificate has the correct issuing permissions, False otherwise.
        """
        cert_issuer = CertificateIssuer(self.coder, self.backend)
        return cert_issuer.check_issuer_has_subject_permissions(certificate, issuer)

    def verify_issued_certificate(
        self, certificate: Certificate, issuer: Certificate
    ) -> bool:
        """
        Verifies if a certificate is issued and verified by it's issuer.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be verified.
        issuer : Certificate
            The issuer of the certificate.

        Returns
        -------
        bool
            True if the certificate is valid, False otherwise.
        """
        if (
            self.certificate_is_issued(certificate)
            and self.check_corresponding_issuer(certificate, issuer)
            and self.verify_correct_issuing_permissions(certificate, issuer)
        ):
            if self.signature_is_nist_p256(
                certificate
            ) and self.verification_key_is_nist_p256(certificate):
                if self.verify_signature(
                    certificate.certificate["toBeSigned"],
                    certificate.certificate["signature"],
                    issuer.certificate["toBeSigned"]["verifyKeyIndicator"][1],
                ):
                    certificate.verified = True
                    return True
        return False

    def verify(self, certificate: Certificate, issuer: Certificate = None) -> bool:
        """
        Verify the certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be verified.
        issuer : Certificate
            The issuer of the certificate.

        Returns
        -------
        bool
            True if the certificate is valid, False otherwise.
        """
        if issuer is not None and self.certificate_is_issued(certificate):
            return self.verify_issued_certificate(certificate, issuer)
        if self.certificate_is_self_signed(certificate):
            return self.verify_self_signed_certificate(certificate)
        return False


class CertificateIssuer:
    """
    Class that issues certificates.
    """

    def __init__(self, coder: SecurityCoder, backend: ECDSABackend):
        """
        Initialize the Certificate Issuer.

        Parameters
        ----------
        coder : SecurityCoder
            Security Coder.
        backend : ECDSABackend
            ECDSA Backend.
        """
        self.coder: SecurityCoder = coder
        self.backend: ECDSABackend = backend

    def certificate_is_self_signed(
        self, certificate: Certificate, issuer: OwnCertificate
    ) -> bool:
        """
        Check if the certificate is self signed.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.
        issuer : OwnCertificate
            The issuer of the certificate.

        Returns
        -------
        bool
            True if the certificate is self signed, False otherwise.
        """
        return certificate == issuer

    def set_issuer_as_self(self, certificate: Certificate) -> None:
        """
        Set the issuer as self.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be set.
        """
        certificate.certificate["issuer"] = ("self", "sha256")

    def set_issuer(self, certificate: Certificate, issuer: OwnCertificate) -> None:
        """
        Set the issuer of the certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be set.
        issuer : OwnCertificate
            The issuer of the certificate.
        """
        certificate.certificate["issuer"] = ("sha256AndDigest", issuer.as_hashedid8())

    def check_enough_min_chain_length_for_issuer(self, issuer: OwnCertificate) -> bool:
        """
        Checks if the chain of trust is lengthy enough.

        Parameters
        ----------
        issuer : OwnCertificate
            The issuer of the certificate.

        Returns
        -------
        bool
            True if the chain of trust is lengthy enough, False otherwise.
        """
        issuer_permissions = issuer.certificate["toBeSigned"]["certIssuePermissions"]
        if not any(
            permission["minChainLength"] < 1 for permission in issuer_permissions
        ):
            return True
        return False
        # for permission in issuer_permissions:
        #     if permission['minChainLength'] < 1:
        #         return False
        # return True

    def get_list_of_psid_from_cert_issue_permissions(
        self, certificate: Certificate
    ) -> list[int]:
        """
        Get the list of PSID from the certificate issue permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to get the PSID from.

        Returns
        -------
        list[int]
            The list of PSID.
        """
        to_return = []
        if self.certificate_wants_cert_issue_permissions(certificate):
            cert_issue_permissions = certificate.certificate["toBeSigned"][
                "certIssuePermissions"
            ]
            for permission in cert_issue_permissions:
                if permission["subjectPermissions"][0] == "explicit":
                    for elem in permission["subjectPermissions"][1]:
                        to_return.append(elem["psid"])
        return to_return

    def get_list_of_psid_from_app_permissions(
        self, certificate: Certificate
    ) -> list[int]:
        """
        Get the list of PSID from the application permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to get the PSID from.

        Returns
        -------
        list[int]
            The list of PSID.
        """
        cert_app_permissions = certificate.certificate["toBeSigned"]["appPermissions"]
        to_return = []
        for elem in cert_app_permissions:
            to_return.append(elem["psid"])
        return to_return

    def get_list_of_needed_permissions(self, certificate: Certificate) -> list[int]:
        """
        Gets the list of needed permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to get the needed permissions from.

        Returns
        -------
        list[int]
            The list of needed permissions.
        """
        to_return = self.get_list_of_psid_from_cert_issue_permissions(certificate)
        to_return.extend(self.get_list_of_psid_from_app_permissions(certificate))
        to_return = list(dict.fromkeys(to_return))
        return to_return

    def get_list_of_allowed_persmissions(self, issuer: OwnCertificate) -> list[int]:
        """
        Gets the list of allowed permissions.

        Parameters
        ----------
        issuer : OwnCertificate
            The issuer of the certificate.

        Returns
        -------
        list[int]
            The list of allowed permissions.
        """
        to_return = []
        issuer_permissions = []
        if self.certificate_wants_cert_issue_permissions(issuer):
            cert_issue_permissions = issuer.certificate["toBeSigned"][
                "certIssuePermissions"
            ]
            for permission in cert_issue_permissions:
                issuer_permissions = permission["subjectPermissions"]
                if issuer_permissions[0] == "explicit":
                    issuer_permissions = issuer_permissions[1]
                    for elem in issuer_permissions:
                        to_return.append(elem["psid"])
        return to_return

    def check_all_requested_permissions_are_allowed(
        self, certificate_permissions: list[int], issuer_permissions: list[int]
    ) -> bool:
        """
        Check if all the requested permissions are allowed.

        Parameters
        ----------
        certificate_permissions : list[int]
            The requested permissions.
        issuer_permissions : list[int]
            The allowed permissions.

        Returns
        -------
        bool
            True if all the requested permissions are allowed, False otherwise.
        """
        return all(item in issuer_permissions for item in certificate_permissions)

    def certificate_has_all_permissions(self, certificate: Certificate) -> bool:
        """
        Check if the certificate has all permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the certificate has all permissions, False otherwise.
        """
        if "certIssuePermissions" in certificate.certificate["toBeSigned"]:
            cert_issue_permissions: list = certificate.certificate["toBeSigned"][
                "certIssuePermissions"
            ]
            for permission in cert_issue_permissions:
                if permission["subjectPermissions"][0] == "all":
                    return True
        return False

    def check_issuer_has_subject_permissions(
        self, certificate: Certificate, issuer: Certificate
    ) -> bool:
        """
        Check if the issuer has the subject permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.
        issuer : Certificate
            The issuer of the certificate.

        Returns
        -------
        bool
            True if the issuer has the subject permissions, False otherwise.
        """
        if self.certificate_has_all_permissions(issuer):
            return True
        return self.check_all_requested_permissions_are_allowed(
            self.get_list_of_needed_permissions(certificate),
            self.get_list_of_allowed_persmissions(issuer),
        )

    def certificate_wants_cert_issue_permissions(
        self, certificate: Certificate
    ) -> bool:
        """
        Check if the certificate wants certificate issue permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to check.

        Returns
        -------
        bool
            True if the certificate wants certificate issue permissions, False otherwise.
        """
        if "certIssuePermissions" in certificate.certificate["toBeSigned"]:
            return True
        return False

    def set_chain_length_issue_permissions(
        self, certificate: Certificate, issuer: OwnCertificate
    ) -> None:
        """
        Set the chain length issue permissions.

        Parameters
        ----------
        certificate : Certificate
            The certificate to set.
        issuer : OwnCertificate
            The issuer of the certificate.
        """
        if self.certificate_wants_cert_issue_permissions(certificate):
            needed_issuing_permissions_capability = (
                self.get_list_of_psid_from_cert_issue_permissions(certificate)
            )
            if self.certificate_has_all_permissions(issuer):
                # Get the "all" permissions
                all_permissions = {}
                for permission in issuer.certificate["toBeSigned"][
                    "certIssuePermissions"
                ]:
                    if permission["subjectPermissions"][0] == "all":
                        all_permissions = permission
                for permission in certificate.certificate["toBeSigned"][
                    "certIssuePermissions"
                ]:
                    permission["minChainLength"] = copy(
                        all_permissions["minChainLength"]
                    )
            else:
                certificate.certificate["toBeSigned"]["certIssuePermissions"] = []
                for permission in issuer.certificate["toBeSigned"][
                    "certIssuePermissions"
                ]:
                    if permission["subjectPermissions"][0] == "explicit":
                        for elem in permission["subjectPermissions"][1]:
                            if elem["psid"] in needed_issuing_permissions_capability:
                                certificate.certificate["toBeSigned"][
                                    "certIssuePermissions"
                                ].append(deepcopy(permission))
            for permission in certificate.certificate["toBeSigned"][
                "certIssuePermissions"
            ]:
                permission["minChainLength"] -= 1
            for permission in list(
                certificate.certificate["toBeSigned"]["certIssuePermissions"]
            ):
                if permission["minChainLength"] < 1:
                    certificate.certificate["toBeSigned"][
                        "certIssuePermissions"
                    ].remove(permission)

    def sign_certificate(
        self, certificate: Certificate, issuer: OwnCertificate
    ) -> None:
        """
        Sign the certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be signed.
        issuer : OwnCertificate
            The issuer of the certificate.
        """
        certificate.certificate["signature"] = self.backend.sign(
            self.coder.encode_ToBeSignedCertificate(
                certificate.certificate["toBeSigned"]
            ),
            issuer.key_id,
        )
        certificate.issuer = issuer

    def issue_certificate(
        self, certificate: Certificate, issuer: OwnCertificate
    ) -> None:
        """
        Issue a certificate.

        Parameters
        ----------
        certificate : Certificate
            The certificate to be issued.
        issuer : OwnCertificate
            The issuer of the certificate.
        """
        if self.certificate_is_self_signed(certificate, issuer):
            self.set_issuer_as_self(certificate)
            self.sign_certificate(certificate, certificate)
        elif self.check_issuer_has_subject_permissions(
            certificate, issuer
        ) and self.check_enough_min_chain_length_for_issuer(issuer):
            self.set_chain_length_issue_permissions(certificate, issuer)
            self.set_issuer(certificate, issuer)
            self.sign_certificate(certificate, issuer)



"""
Cooperatie Awareness Basic Service

This file contains the class for the Cooperative Awareness Basic Service.
"""

import logging
from ..local_dynamic_map.ldm_facility import LDMFacility
from ..local_dynamic_map.ldm_constants import CAM
from .cam_transmission_management import (
    CAMTransmissionManagement,
    VehicleData,
)
from ...btp.router import Router as BTPRouter
from .cam_coder import CAMCoder
from .cam_reception_management import (
    CAMReceptionManagement,
)
from .cam_ldm_adaptation import CABasicServiceLDM


class CooperativeAwarenessBasicService:
    """
    Cooperative Awareness Basic Service

    Attributes
    ----------
    btp_router : BTPRouter
        BTP Router.
    cam_coder : CAMCoder
        CAM Coder.
    vehicle_data : VehicleData
        Vehicle Data.
    cam_transmission_management : CAMTransmissionManagement
        CAM Transmission Management.
    cam_reception_management : CAMReceptionManagement
        CAM Reception Management.
    """

    def __init__(
        self,
        btp_router: BTPRouter,
        vehicle_data: VehicleData,
        ldm: LDMFacility = None,
    ) -> None:
        """
        Initialize the Cooperative Awareness Basic Service.

        Parameters
        ----------
        btp_router : BTPRouter
            BTP Router.
        vehicle_data : VehicleData
            Vehicle Data.
        ldm: LDMFacility
            Local Dynamic Map (LDM) Service.
        """
        self.logging = logging.getLogger("ca_basic_service")

        self.btp_router = btp_router
        self.cam_coder = CAMCoder()
        self.vehicle_data = vehicle_data
        ca_basic_service_ldm = None
        if ldm is not None:
            ca_basic_service_ldm = CABasicServiceLDM(ldm, [CAM], 5)
        self.cam_transmission_management = CAMTransmissionManagement(
            btp_router=btp_router,
            cam_coder=self.cam_coder,
            vehicle_data=self.vehicle_data,
            ca_basic_service_ldm=ca_basic_service_ldm,
        )
        self.cam_reception_management = CAMReceptionManagement(
            cam_coder=self.cam_coder,
            btp_router=self.btp_router,
            ca_basic_service_ldm=ca_basic_service_ldm,
        )

        self.logging.info("CA Basic Service Started!")



import asn1tools
from .asn1.denm_asn1 import DENM_ASN1_DESCRIPTIONS


class DENMCoder:
    """
    Class for encoding/decoding Decentralized Environment Notification Messages (DENM)

    Attributes
    ----------
    asn_coder : asn1tools.Coder
        ASN.1 coder.
    """

    def __init__(self) -> None:
        """
        Initialize the DENM Coder.
        """
        self.asn_coder = asn1tools.compile_string(
            DENM_ASN1_DESCRIPTIONS, codec='uper')

    def encode(self, denm: dict) -> bytes:
        """
        Encodes a DENM message.

        Parameters
        ----------
        denm : dict
            DENM message.

        Returns
        -------
        bytes
            Encoded DENM message.
        """
        return self.asn_coder.encode('DENM', denm)

    def decode(self, denm: bytes) -> dict:
        """
        Decodes a DENM message.

        Parameters
        ----------
        denm : bytes
            Encoded DENM message.

        Returns
        -------
        dict
            DENM message.
        """
        return self.asn_coder.decode('DENM', denm)



import logging

from .ldm_classes import Location

from .ldm_maintenance_reactive import LDMMaintenanceReactive
from .ldm_maintenance_thread import LDMMaintenanceThread

from .ldm_service_reactive import LDMServiceReactive
from .ldm_service_threads import LDMServiceThreads

from .dictionary_database import DictionaryDataBase
from .tinydb_database import TinyDB

from .ldm_facility import LDMFacility

from .exceptions import LDMMaintenanceKeyError, LDMServiceKeyError, LDMDatabaseKeyError


def ldm_factory(
    ldm_location: Location,
    ldm_maintenance_type: str = "Reactive",
    ldm_service_type: str = "Reactive",
    ldm_database_type: str = "Dictionary",
) -> LDMFacility:
    """
    Factory function to create a Local Dynamic Map Facility.

    Parameters
    ----------
    ldm_location : Location
        Location object to be used by the LDM Facility.
    ldm_maintenance_type : str, optional
        Type of LDM Maintenance to be used. Defaults to "Reactive".
    ldm_service_type : str, optional
        Type of LDM Service to be used. Defaults to "Reactive".
    ldm_database_type : str, optional
        Type of LDM Database to be used. Defaults to "Dictionary".
    """

    logs = logging.getLogger("local_dynamic_map")

    if ldm_database_type == "Dictionary":
        ldm_database = DictionaryDataBase()
        logs.info('LDM Database "Dictionary" configured.')
    elif ldm_database_type == "TinyDB":
        ldm_database = TinyDB()
        logs.info('LDM Database "TindyDB" configured.')
    else:
        logs.error(
            'LDM Database must be either "Dictionary" or "TindyDB". %s is an invalid type for LDM Database.',
            ldm_database_type,
        )
        raise LDMDatabaseKeyError(
            f'LDM Database must be either "Dictionary" or "TindyDB". {ldm_database_type} is an invalid type.'
        )

    if ldm_maintenance_type == "Reactive":
        ldm_maintenance = LDMMaintenanceReactive(ldm_location, ldm_database)
        logs.info('LDM Maintenance "Reactive" configured.')
    elif ldm_maintenance_type == "Thread":
        ldm_maintenance = LDMMaintenanceThread(ldm_location, ldm_database, None)
        logs.info('LDM Maintenance "Thread" configured.')
    else:
        logs.error(
            'LDM Maintenance must be either "Reactive" or "Thread". %s is an invalid type for LDM Maintenance.',
            ldm_maintenance_type,
        )
        raise LDMMaintenanceKeyError(
            f'LDM Maintenance must be either "Reactive" or "Thread". {ldm_maintenance_type} is an invalid type.'
        )

    if ldm_service_type == "Reactive":
        ldm_service = LDMServiceReactive(ldm_maintenance)
        logs.info('LDM Service "Reactive" configured.')
    elif ldm_service_type == "Thread":
        ldm_service = LDMServiceThreads(ldm_maintenance)
        logs.info('LDM Service "Thread" configured.')
    else:
        logs.error(
            'LDM Service must be either "Reactive" or "Thread". %s is an invalid type for LDM Service.',
            ldm_service_type,
        )
        raise LDMServiceKeyError(
            f'LDM Service must be either "Reactive" or "Thread". {ldm_service_type} is an invalid type for LDM Service.'
        )

    ldm_facility = LDMFacility(ldm_maintenance, ldm_service)
    logs.info(
        'LDM Facility configured with: LDM Maintenance: "%s", LDM Service: "%s", LDM Database: "%s".',
        ldm_maintenance_type,
        ldm_service_type,
        ldm_database_type,
    )
    return ldm_facility



import logging
from .denm_coder import DENMCoder
from ...btp.service_access_point import BTPDataIndication
from ...btp.router import Router as BTPRouter

from ..local_dynamic_map.ldm_facility import LDMFacility
from ..local_dynamic_map.ldm_classes import (
    RegisterDataProviderReq,
    AddDataProviderReq,
    TimeValidity,
    TimestampIts,
    Location,
)
from ..local_dynamic_map.ldm_constants import DENM


class DENMReceptionManagement:
    """
    This class is responsible for the DENM reception management.

    Attributes
    ----------
    denm_coder : DENMCoder
        DENM Coder object.

    """

    def __init__(
        self, denm_coder: DENMCoder, btp_router: BTPRouter, ldm: LDMFacility
    ) -> None:
        """
        Initialize the DENM Reception Management.

        Parameters
        ----------
        denm_coder : DENMCoder
            DENM Coder object.
        """

        self.logging = logging.getLogger("denm_service")

        self.denm_coder = denm_coder
        self.btp_router = btp_router
        self.btp_router.register_indication_callback_btp(
            port=2002, callback=self.reception_callback
        )
        self.ldm_facility = ldm
        if self.ldm_facility is not None:
            self.ldm_facility.if_ldm_3.register_data_provider(
                RegisterDataProviderReq(
                    application_id=DENM,
                    access_permissions=[
                        DENM,
                    ],
                    time_validity=TimeValidity(1000),
                )
            )

    def feed_ldm(self, denm: dict) -> None:
        """
        Send DENM to LDM Facility.

        Parameters
        ----------
        denm : dict
            DENM message.
        """
        if self.ldm_facility is not None:
            data = AddDataProviderReq(
                application_id=DENM,
                time_stamp=TimestampIts(
                    denm["denm"]["management"]["detectionTime"] % 65536
                ),
                location=Location.location_builder_circle(
                    latitude=denm["denm"]["management"]["eventPosition"]["latitude"],
                    longitude=denm["denm"]["management"]["eventPosition"]["longitude"],
                    altitude=denm["denm"]["management"]["eventPosition"]["altitude"][
                        "altitudeValue"
                    ],
                    radius=0,
                ),
                data_object=denm,
                time_validity=TimeValidity(3),
            )
            self.ldm_facility.if_ldm_3.add_provider_data(data)
            # if not isinstance(response.dataObjectID, int):
            #     raise response.dataObjectID
            self.logging.debug("Added DENM data to the LDM. Data: %s", data)

    def reception_callback(self, btp_indication: BTPDataIndication) -> None:
        """
        Callback for the reception of a DENM.

        Parameters
        ----------
        btp_indication : BTPDataIndication
            BTP Data Indication.
        """
        denm = self.denm_coder.decode(btp_indication.data)
        self.feed_ldm(denm)
        self.logging.debug(
            "Received DENM with timestamp: %s, station_id: %s",
            denm["denm"]["management"]["referenceTime"],
            denm["header"]["stationId"],
        )
